# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)
# ----------------------------------------------------------------------------

"""
Tests for the Variable Explorer Collections Editor.
"""

# Standard library imports
import os  # Example module for testing display inside CollecitonsEditor
from os import path
import copy
import datetime
from xml.dom.minidom import parseString
from unittest.mock import Mock, patch

# Third party imports
import numpy
from packaging.version import parse
import pandas
import pytest
from flaky import flaky
from qtpy.QtCore import Qt, QPoint
from qtpy.QtWidgets import QDateEdit, QMessageBox, QWidget

# Local imports
from spyder.config.manager import CONF
from spyder.widgets.collectionseditor import (
    CollectionsEditor, CollectionsEditorTableView, CollectionsEditorWidget,
    CollectionsModel, LARGE_NROWS, natsort, RemoteCollectionsEditorTableView,
    ROWS_TO_LOAD)
from spyder.plugins.variableexplorer.widgets.collectionsdelegate import (
    SELECT_ROW_BUTTON_SIZE
)
from spyder.plugins.variableexplorer.widgets.tests.test_dataframeeditor import (
    generate_pandas_indexes)
from spyder_kernels.utils.nsview import get_size


# =============================================================================
# Constants
# =============================================================================
# Full path to this file's parent directory for loading data
LOCATION = path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


# =============================================================================
# Utility functions and classes
# =============================================================================
def data(cm, i, j):
    return cm.data(cm.index(i, j))


def data_table(cm, n_rows, n_cols):
    return [[data(cm, i, j) for i in range(n_rows)] for j in range(n_cols)]


class MockParent(QWidget):

    def __init__(self):
        QWidget.__init__(self)
        self.proxy_model = None


# =============================================================================
# Pytest Fixtures
# =============================================================================
@pytest.fixture
def nonsettable_objects_data():
    """Rturn Python objects with immutable attribs to test CollectionEditor."""
    test_objs = [pandas.Period("2018-03"), pandas.Categorical([1, 2, 42])]
    expected_objs = [pandas.Period("2018-03"), pandas.Categorical([1, 2, 42])]
    keys_test = [["_typ", "day", "dayofyear", "hour"],
                 ["_typ", "nbytes", "ndim"]]
    return zip(test_objs, expected_objs, keys_test)


# =============================================================================
# Tests
# ============================================================================
def test_rename_variable(qtbot):
    """Test renaming of the correct variable."""
    variables = {'a': 1,
                 'b': 2,
                 'c': 3,
                 'd': '4',
                 'e': 5}
    editor = CollectionsEditorTableView(None, variables.copy())
    qtbot.addWidget(editor)
    editor.setCurrentIndex(editor.model.index(1, 0))

    editor.rename_item(new_name='b2')
    assert editor.model.rowCount() == 5
    assert data(editor.model, 0, 0) == 'a'
    assert data(editor.model, 1, 0) == 'b2'
    assert data(editor.model, 2, 0) == 'c'
    assert data(editor.model, 3, 0) == 'd'
    assert data(editor.model, 4, 0) == 'e'

    # Reset variables and try renaming one again
    new_variables = {'a': 1,
                     'b': 2,
                     'b2': 2,
                     'c': 3,
                     'd': '4',
                     'e': 5}
    editor.set_data(new_variables.copy())
    editor.adjust_columns()
    editor.setCurrentIndex(editor.model.index(1, 0))
    editor.rename_item(new_name='b3')
    assert editor.model.rowCount() == 6
    assert data(editor.model, 0, 0) == 'a'
    assert data(editor.model, 1, 0) == 'b2'
    assert data(editor.model, 2, 0) == 'b3'
    assert data(editor.model, 3, 0) == 'c'
    assert data(editor.model, 4, 0) == 'd'
    assert data(editor.model, 5, 0) == 'e'


def test_remove_variable(qtbot):
    """Test removing of the correct variable."""
    variables = {'a': 1,
                 'b': 2,
                 'c': 3,
                 'd': '4',
                 'e': 5}
    editor = CollectionsEditorTableView(None, variables.copy())
    qtbot.addWidget(editor)
    editor.setCurrentIndex(editor.model.index(1, 0))

    editor.remove_item(force=True)
    assert editor.model.rowCount() == 4
    assert data(editor.model, 0, 0) == 'a'
    assert data(editor.model, 1, 0) == 'c'
    assert data(editor.model, 2, 0) == 'd'
    assert data(editor.model, 3, 0) == 'e'

    # Reset variables and try removing one again
    editor.set_data(variables.copy())
    editor.adjust_columns()
    editor.setCurrentIndex(editor.model.index(1, 0))
    editor.remove_item(force=True)
    assert editor.model.rowCount() == 4
    assert data(editor.model, 0, 0) == 'a'
    assert data(editor.model, 1, 0) == 'c'
    assert data(editor.model, 2, 0) == 'd'
    assert data(editor.model, 3, 0) == 'e'


def test_remove_remote_variable(qtbot, monkeypatch):
    """Test the removing of the correct remote variable."""
    variables = {'a': {'type': 'int',
                       'size': 1,
                       'view': '1',
                       'python_type': 'int',
                       'numpy_type': 'Unknown'},
                 'b': {'type': 'int',
                       'size': 1,
                       'view': '2',
                       'python_type': 'int',
                       'numpy_type': 'Unknown'},
                 'c': {'type': 'int',
                       'size': 1,
                       'view': '3',
                       'python_type': 'int',
                       'numpy_type': 'Unknown'},
                 'd': {'type': 'str',
                       'size': 1,
                       'view': '4',
                       'python_type': 'int',
                       'numpy_type': 'Unknown'},
                 'e': {'type': 'int',
                       'size': 1,
                       'view': '5',
                       'python_type': 'int',
                       'numpy_type': 'Unknown'}}
    editor = RemoteCollectionsEditorTableView(None, variables.copy())
    qtbot.addWidget(editor)
    editor.setCurrentIndex(editor.model.index(1, 0))

    # Monkey patch remove variables
    def remove_values(ins, names):
        assert names == ['b']
        data = {'a': {'type': 'int',
                      'size': 1,
                      'view': '1',
                      'python_type': 'int',
                      'numpy_type': 'Unknown'},
                'c': {'type': 'int',
                      'size': 1,
                      'view': '3',
                      'python_type': 'int',
                      'numpy_type': 'Unknown'},
                'd': {'type': 'str',
                      'size': 1,
                      'view': '4',
                      'python_type': 'int',
                      'numpy_type': 'Unknown'},
                'e': {'type': 'int',
                      'size': 1,
                      'view': '5',
                      'python_type': 'int',
                      'numpy_type': 'Unknown'}}
        editor.set_data(data)
    monkeypatch.setattr(
        'spyder.widgets'
        '.collectionseditor.RemoteCollectionsEditorTableView.remove_values',
        remove_values)

    editor.remove_item(force=True)
    assert editor.model.rowCount() == 4
    assert data(editor.model, 0, 0) == 'a'
    assert data(editor.model, 1, 0) == 'c'
    assert data(editor.model, 2, 0) == 'd'
    assert data(editor.model, 3, 0) == 'e'

    # Reset variables and try removing one again
    editor.set_data(variables.copy())
    editor.adjust_columns()
    editor.setCurrentIndex(editor.model.index(1, 0))
    editor.remove_item(force=True)
    assert editor.model.rowCount() == 4
    assert data(editor.model, 0, 0) == 'a'
    assert data(editor.model, 1, 0) == 'c'
    assert data(editor.model, 2, 0) == 'd'
    assert data(editor.model, 3, 0) == 'e'


def test_filter_rows(qtbot):
    """Test rows filtering."""
    data = (
        {'dfa':
            {'type': 'DataFrame',
             'size': (2, 1),
             'view': 'Column names: 0',
             'python_type': 'DataFrame',
             'numpy_type': 'Unknown'},
         'dfb':
            {'type': 'DataFrame',
             'size': (2, 1),
             'view': 'Column names: 0',
             'python_type': 'DataFrame',
             'numpy_type': 'Unknown'}}
    )
    editor = RemoteCollectionsEditorTableView(None, data)
    qtbot.addWidget(editor)

    # Initially two rows
    assert editor.model.rowCount() == 2

    # Match two rows by name
    editor.do_find("df")
    assert editor.model.rowCount() == 2

    # Match two rows by type
    editor.do_find("DataFrame")
    assert editor.model.rowCount() == 2

    # Only one match
    editor.do_find("dfb")
    assert editor.model.rowCount() == 1

    # No match
    editor.do_find("dfbc")
    assert editor.model.rowCount() == 0


def test_remote_make_data_function():
    """
    Test that the function returned by make_data_function() is the expected
    one.
    """
    variables = {'a': {'type': 'int',
                       'size': 1,
                       'view': '1',
                       'python_type': 'int',
                       'numpy_type': 'Unknown'}}
    mock_shellwidget = Mock()
    editor = RemoteCollectionsEditorTableView(
        None, variables, mock_shellwidget
    )
    index = editor.model.index(0, 0)
    data_function = editor.delegate.make_data_function(index)
    value = data_function()
    mock_shellwidget.get_value.assert_called_once_with('a')
    assert value == mock_shellwidget.get_value.return_value


def test_create_dataframeeditor_with_correct_format(qtbot):
    df = pandas.DataFrame(['foo', 'bar'])
    editor = CollectionsEditorTableView(None, {'df': df})
    qtbot.addWidget(editor)
    CONF.set('variable_explorer', 'dataframe_format', '10d')
    editor.delegate.createEditor(None, None, editor.model.index(0, 3))
    dataframe_editor = next(iter(editor.delegate._editors.values()))['editor']
    qtbot.addWidget(dataframe_editor)
    dataframe_editor.dataModel._format_spec == '10d'


def test_collectionsmodel_with_two_ints():
    coll = {'x': 1, 'y': 2}
    cm = CollectionsModel(MockParent(), coll)

    assert cm.rowCount() == 2
    assert cm.columnCount() == 4
    # dict is unordered, so first row might be x or y
    assert data(cm, 0, 0) in {'x',
                              'y'}
    if data(cm, 0, 0) == 'x':
        row_with_x = 0
        row_with_y = 1
    else:
        row_with_x = 1
        row_with_y = 0
    assert data(cm, row_with_x, 1) == 'int'
    assert data(cm, row_with_x, 2) == 1
    assert data(cm, row_with_x, 3) == '1'
    assert data(cm, row_with_y, 0) == 'y'
    assert data(cm, row_with_y, 1) == 'int'
    assert data(cm, row_with_y, 2) == 1
    assert data(cm, row_with_y, 3) == '2'


def test_collectionsmodel_with_index():
    # Regression test for spyder-ide/spyder#3380,
    # modified for spyder-ide/spyder#3758.
    for rng_name, rng in generate_pandas_indexes().items():
        coll = {'rng': rng}
        cm = CollectionsModel(MockParent(), coll)
        assert data(cm, 0, 0) == 'rng'
        assert data(cm, 0, 1) == rng_name
        assert data(cm, 0, 2) == '(20,)' or data(cm, 0, 2) == '(20L,)'
    try:
        assert data(cm, 0, 3) == rng._summary()
    except AttributeError:
        assert data(cm, 0, 3) == rng.summary()


def test_shows_dataframeeditor_when_editing_index(monkeypatch):
    for __, rng in generate_pandas_indexes().items():
        MockDataFrameEditor = Mock()
        mockDataFrameEditor_instance = MockDataFrameEditor()
        attr_to_patch_dfedit = ('spyder.plugins.variableexplorer.widgets.' +
                                'dataframeeditor.DataFrameEditor')
        monkeypatch.setattr(attr_to_patch_dfedit, MockDataFrameEditor)
        coll = {'rng': rng}
        editor = CollectionsEditorTableView(None, coll)
        editor.delegate.createEditor(None, None,
                                     editor.model.index(0, 3))
        mockDataFrameEditor_instance.show.assert_called_once_with()


def test_sort_numpy_numeric_collectionsmodel():
    if parse(numpy.__version__) >= parse("2.0.0"):
        np20 = True
    else:
        np20 = False

    var_list = [
        numpy.float64(1e16),
        numpy.float64(10),
        numpy.float64(1),
        numpy.float64(0.1),
        numpy.float64(1e-6),
        numpy.float64(0),
        numpy.float64(-1e-6),
        numpy.float64(-1),
        numpy.float64(-10),
        numpy.float64(-1e16),
    ]
    cm = CollectionsModel(MockParent(), var_list)
    assert cm.rowCount() == 10
    assert cm.columnCount() == 4

    # Sort by index
    cm.sort(0)
    assert data_table(cm, 10, 4) == [
        list(range(0, 10)),
        ["float64"] * 10,
        [1] * 10,
        [
            "np.float64(1e+16)" if np20 else "1e+16",
            "np.float64(10.0)" if np20 else "10.0",
            "np.float64(1.0)" if np20 else "1.0",
            "np.float64(0.1)" if np20 else "0.1",
            "np.float64(1e-06)" if np20 else "1e-06",
            "np.float64(0.0)" if np20 else "0.0",
            "np.float64(-1e-06)" if np20 else "-1e-06",
            "np.float64(-1.0)" if np20 else "-1.0",
            "np.float64(-10.0)" if np20 else "-10.0",
            "np.float64(-1e+16)" if np20 else "-1e+16",
        ],
    ]

    # Sort by value
    cm.sort(3)
    assert data_table(cm, 10, 4) == [
        list(range(9, -1, -1)),
        ["float64"] * 10,
        [1] * 10,
        [
            "np.float64(-1e+16)" if np20 else "-1e+16",
            "np.float64(-10.0)" if np20 else "-10.0",
            "np.float64(-1.0)" if np20 else "-1.0",
            "np.float64(-1e-06)" if np20 else "-1e-06",
            "np.float64(0.0)" if np20 else "0.0",
            "np.float64(1e-06)" if np20 else "1e-06",
            "np.float64(0.1)" if np20 else "0.1",
            "np.float64(1.0)" if np20 else "1.0",
            "np.float64(10.0)" if np20 else "10.0",
            "np.float64(1e+16)" if np20 else "1e+16",
        ],
    ]


def test_sort_float_collectionsmodel():
    var_list = [
        float(1e16), float(10), float(1), float(0.1), float(1e-6),
        float(0), float(-1e-6), float(-1), float(-10), float(-1e16)
        ]
    cm = CollectionsModel(MockParent(), var_list)
    assert cm.rowCount() == 10
    assert cm.columnCount() == 4
    cm.sort(0)  # sort by index
    assert data_table(cm, 10, 4) == [list(range(0, 10)),
                                     [u'float']*10,
                                     [1]*10,
                                     ['1e+16', '10.0', '1.0', '0.1',
                                      '1e-06', '0.0', '-1e-06',
                                      '-1.0', '-10.0', '-1e+16']]
    cm.sort(3)  # sort by value
    assert data_table(cm, 10, 4) == [list(range(9, -1, -1)),
                                     [u'float']*10,
                                     [1]*10,
                                     ['-1e+16', '-10.0', '-1.0',
                                      '-1e-06', '0.0', '1e-06',
                                      '0.1', '1.0', '10.0', '1e+16']]


def test_sort_collectionsmodel():
    var_list1 = [0, 1, 2]
    var_list2 = [3, 4, 5, 6]
    var_dataframe1 = pandas.DataFrame([[1, 2, 3], [20, 30, 40], [2, 2, 2]])
    var_dataframe2 = pandas.DataFrame([[1, 2, 3], [20, 30, 40]])
    var_series1 = pandas.Series(var_list1)
    var_series2 = pandas.Series(var_list2)

    coll = [1, 3, 2]
    cm = CollectionsModel(MockParent(), coll)
    assert cm.rowCount() == 3
    assert cm.columnCount() == 4
    cm.sort(0)  # sort by index
    assert data_table(cm, 3, 4) == [[0, 1, 2],
                                    ['int', 'int', 'int'],
                                    [1, 1, 1],
                                    ['1', '3', '2']]
    cm.sort(3)  # sort by value
    assert data_table(cm, 3, 4) == [[0, 2, 1],
                                    ['int', 'int', 'int'],
                                    [1, 1, 1],
                                    ['1', '2', '3']]

    coll = [1, var_list1, var_list2, var_dataframe1, var_dataframe2,
            var_series1, var_series2]
    cm = CollectionsModel(MockParent(), coll)
    assert cm.rowCount() == 7
    assert cm.columnCount() == 4

    cm.sort(1)  # sort by type
    assert data_table(cm, 7, 4) == [
        [3, 4, 5, 6, 0, 1, 2],
        ['DataFrame', 'DataFrame', 'Series', 'Series', 'int', 'list', 'list'],
        ['(3, 3)', '(2, 3)', '(3,)', '(4,)', 1, 3, 4],
        ['Column names: 0, 1, 2',
         'Column names: 0, 1, 2',
         'Series object of pandas.core.series module',
         'Series object of pandas.core.series module',
         '1',
         '[0, 1, 2]',
         '[3, 4, 5, 6]']]

    cm.sort(2)  # sort by size
    assert data_table(cm, 7, 4) == [
        [3, 4, 5, 6, 0, 1, 2],
        ['DataFrame', 'DataFrame', 'Series', 'Series', 'int', 'list', 'list'],
        ['(2, 3)', '(3,)', '(3, 3)', '(4,)', 1, 3, 4],
        ['Column names: 0, 1, 2',
         'Column names: 0, 1, 2',
         'Series object of pandas.core.series module',
         'Series object of pandas.core.series module',
         '1',
         '[0, 1, 2]',
         '[3, 4, 5, 6]']] or data_table(cm, 7, 4) == [
        [0, 1, 2, 4, 5, 3, 6],
        [u'int', u'list', u'list', u'DataFrame', u'Series', u'DataFrame',
         u'Series'],
        [1, 3, 4, u'(2, 3)', u'(3,)', u'(3, 3)', u'(4,)'],
        ['1',
         '[0, 1, 2]',
         '[3, 4, 5, 6]',
         'Column names: 0, 1, 2',
         'Series object of pandas.core.series module',
         'Column names: 0, 1, 2',
         'Series object of pandas.core.series module',
         ]]


def test_sort_and_fetch_collectionsmodel_with_many_rows():
    coll = list(range(2*LARGE_NROWS))
    cm = CollectionsModel(MockParent(), coll)
    assert cm.rowCount() == cm.rows_loaded == ROWS_TO_LOAD
    assert cm.columnCount() == 4
    cm.sort(1)  # This was causing an issue (#5232)
    cm.fetchMore()
    assert cm.rowCount() == 2 * ROWS_TO_LOAD
    for _ in range(3):
        cm.fetchMore()
    assert cm.rowCount() == len(coll)


def test_rename_and_duplicate_item_in_collection_editor():
    collections = {'list': ([1, 2, 3], False, True),
                   'tuple': ((1, 2, 3), False, False),
                   'dict': ({'a': 1, 'b': 2}, True, True)}
    for coll, rename_enabled, duplicate_enabled in collections.values():
        coll_copy = copy.copy(coll)
        editor = CollectionsEditorTableView(None, coll)
        assert editor.rename_action.isEnabled()
        assert editor.duplicate_action.isEnabled()
        editor.setCurrentIndex(editor.model.index(0, 0))
        editor.refresh_menu()
        assert editor.rename_action.isEnabled() == rename_enabled
        assert editor.duplicate_action.isEnabled() == duplicate_enabled
        if isinstance(coll, list):
            editor.duplicate_item()
            assert editor.source_model.get_data() == coll_copy + [coll_copy[0]]


def test_collectioneditorwidget_refresh_action_disabled():
    """
    Test that the Refresh button is disabled by default.
    """
    lst = [1, 2, 3, 4]
    widget = CollectionsEditorWidget(None, lst.copy())
    assert not widget.refresh_action.isEnabled()


def test_collectioneditor_refresh():
    """
    Test that after pressing the refresh button, the value of the editor is
    replaced by the return value of the data_function.
    """
    old_list = [1, 2, 3, 4]
    new_list = [3, 1, 4, 1, 5]
    editor = CollectionsEditor(None, data_function=lambda: new_list)
    editor.setup(old_list)
    assert editor.get_value() == old_list
    assert editor.widget.refresh_action.isEnabled()
    editor.widget.refresh_action.trigger()
    assert editor.get_value() == new_list


@pytest.mark.parametrize('result', [QMessageBox.Yes, QMessageBox.No])
def test_collectioneditor_refresh_after_edit(result):
    """
    Test that after changing a value in the collections editor, refreshing the
    editor opens a dialog box (which asks for confirmation), and that the
    editor is only refreshed if the user clicks Yes.
    """
    old_list = [1, 2, 3, 4]
    edited_list = [1, 2, 3, 5]
    new_list = [3, 1, 4, 1, 5]
    editor = CollectionsEditor(None, data_function=lambda: new_list)
    editor.setup(old_list)
    editor.show()
    model = editor.widget.editor.source_model
    model.setData(model.index(3, 3), '5')
    with patch('spyder.widgets.collectionseditor.QMessageBox.question',
               return_value=result) as mock_question:
        editor.widget.refresh_action.trigger()
    mock_question.assert_called_once()
    editor.accept()
    if result == QMessageBox.Yes:
        assert editor.get_value() == new_list
    else:
        assert editor.get_value() == edited_list


def test_collectioneditor_refresh_when_variable_deleted(qtbot):
    """
    Test that if the variable is deleted and then the editor is refreshed
    (resulting in data_function raising a KeyError), a critical dialog box
    is displayed and that the editor is closed.
    """
    def datafunc():
        raise KeyError
    lst = [1, 2, 3, 4]
    editor = CollectionsEditor(None, data_function=datafunc)
    editor.setup(lst)
    with patch('spyder.widgets.collectionseditor.QMessageBox'
               '.critical') as mock_critical, \
         qtbot.waitSignal(editor.rejected, timeout=0):
        editor.widget.refresh_action.trigger()
    mock_critical.assert_called_once()


def test_collectioneditor_refresh_nested():
    """
    Open an editor for a list with a tuple nested inside, and then open another
    editor for the nested tuple. Test that refreshing the second editor works.
    """
    old_list = [1, 2, 3, (4, 5)]
    new_list = [1, 2, 3, (4,)]
    editor = CollectionsEditor(None, data_function=lambda: new_list)
    editor.setup(old_list)
    view = editor.widget.editor
    view.edit(view.model.index(3, 3))
    nested_editor = list(view.delegate._editors.values())[0]['editor']
    assert nested_editor.get_value() == (4, 5)
    nested_editor.widget.refresh_action.trigger()
    assert nested_editor.get_value() == (4,)


def test_edit_datetime(monkeypatch):
    """
    Test datetimes are editable and NaT values are correctly handled.

    Regression test for spyder-ide/spyder#13557 and spyder-ide/spyder#8329
    """
    variables = [pandas.NaT, datetime.date.today()]
    editor_list = CollectionsEditorTableView(None, variables)

    # Test that the NaT value cannot be edited on the variable explorer
    editor_list_value = editor_list.delegate.createEditor(
        None, None, editor_list.model.index(0, 3))
    assert editor_list_value is None

    # Test that a date can be edited on the variable explorer
    editor_list_value = editor_list.delegate.createEditor(
        None, None, editor_list.model.index(1, 3))
    assert isinstance(editor_list_value, QDateEdit)


def test_edit_mutable_and_immutable_types(monkeypatch):
    """
    Test that mutable objs/vals are editable in VarExp; immutable ones aren't.

    Regression test for spyder-ide/spyder#5991.
    """
    MockQLineEdit = Mock()
    attr_to_patch_qlineedit = ('spyder.plugins.variableexplorer.widgets.' +
                               'collectionsdelegate.QLineEdit')
    monkeypatch.setattr(attr_to_patch_qlineedit, MockQLineEdit)

    MockTextEditor = Mock()
    attr_to_patch_textedit = ('spyder.plugins.variableexplorer.widgets.' +
                              'collectionsdelegate.TextEditor')
    monkeypatch.setattr(attr_to_patch_textedit, MockTextEditor)

    MockQDateTimeEdit = Mock()
    attr_to_patch_qdatetimeedit = ('spyder.plugins.variableexplorer.widgets.' +
                                   'collectionsdelegate.QDateTimeEdit')
    monkeypatch.setattr(attr_to_patch_qdatetimeedit, MockQDateTimeEdit)

    MockCollectionsEditor = Mock()
    mockCollectionsEditor_instance = MockCollectionsEditor()
    attr_to_patch_coledit = ('spyder.widgets.' +
                             'collectionseditor.CollectionsEditor')
    monkeypatch.setattr(attr_to_patch_coledit, MockCollectionsEditor)

    list_test = [1, "012345678901234567901234567890123456789012",
                 datetime.datetime(2017, 12, 24, 7, 9), [1, 2, 3], (2, "eggs")]
    tup_test = tuple(list_test)

    # Tests for mutable type (list) #
    editor_list = CollectionsEditorTableView(None, list_test)

    # Directly editable values inside list
    editor_list_value = editor_list.delegate.createEditor(
        None, None, editor_list.model.index(0, 3))
    assert editor_list_value is not None
    assert MockQLineEdit.call_count == 1

    # Text Editor for long text inside list
    editor_list.delegate.createEditor(None, None,
                                      editor_list.model.index(1, 3))
    assert MockTextEditor.call_count == 2
    assert not MockTextEditor.call_args[1]["readonly"]

    # Datetime inside list
    editor_list_datetime = editor_list.delegate.createEditor(
        None, None, editor_list.model.index(2, 3))
    assert editor_list_datetime is not None
    assert MockQDateTimeEdit.call_count == 1

    # List inside list
    editor_list.delegate.createEditor(None, None,
                                      editor_list.model.index(3, 3))
    assert mockCollectionsEditor_instance.show.call_count == 1
    assert not mockCollectionsEditor_instance.setup.call_args[1]["readonly"]

    # Tuple inside list
    editor_list.delegate.createEditor(None, None,
                                      editor_list.model.index(4, 3))
    assert mockCollectionsEditor_instance.show.call_count == 2
    assert mockCollectionsEditor_instance.setup.call_args[1]["readonly"]

    # Tests for immutable type (tuple) #
    editor_tup = CollectionsEditorTableView(None, tup_test)

    # Directly editable values inside tuple
    editor_tup_value = editor_tup.delegate.createEditor(
        None, None, editor_tup.model.index(0, 3))
    assert editor_tup_value is None
    assert MockQLineEdit.call_count == 1

    # Text Editor for long text inside tuple
    editor_tup.delegate.createEditor(None, None,
                                     editor_tup.model.index(1, 3))
    assert MockTextEditor.call_count == 4
    assert MockTextEditor.call_args[1]["readonly"]

    # Datetime inside tuple
    editor_tup_datetime = editor_tup.delegate.createEditor(
        None, None, editor_tup.model.index(2, 3))
    assert editor_tup_datetime is None
    assert MockQDateTimeEdit.call_count == 1

    # List inside tuple
    editor_tup.delegate.createEditor(None, None,
                                     editor_tup.model.index(3, 3))
    assert mockCollectionsEditor_instance.show.call_count == 3
    assert mockCollectionsEditor_instance.setup.call_args[1]["readonly"]

    # Tuple inside tuple
    editor_tup.delegate.createEditor(None, None,
                                     editor_tup.model.index(4, 3))
    assert mockCollectionsEditor_instance.show.call_count == 4
    assert mockCollectionsEditor_instance.setup.call_args[1]["readonly"]


@flaky(max_runs=3)
def test_view_module_in_coledit():
    """
    Test that modules don't produce an error when opening in Variable Explorer.

    Also check that they are set as readonly. Regression test for
    spyder-ide/spyder#6080.
    """
    editor = CollectionsEditor()
    editor.setup(os, "module_test", readonly=False)
    assert editor.widget.editor.readonly


def test_notimplementederror_multiindex():
    """
    Test that the NotImplementedError when scrolling a MultiIndex is handled.

    Regression test for spyder-ide/spyder#6284.
    """
    time_deltas = [pandas.Timedelta(minutes=minute)
                   for minute in range(5, 35, 5)]
    time_delta_multiindex = pandas.MultiIndex.from_product([[0, 1, 2, 3, 4],
                                                            time_deltas])
    col_model = CollectionsModel(MockParent(), time_delta_multiindex)
    assert col_model.rowCount() == col_model.rows_loaded == ROWS_TO_LOAD
    assert col_model.columnCount() == 4
    col_model.fetchMore()
    assert col_model.rowCount() == 2 * ROWS_TO_LOAD
    for _ in range(3):
        col_model.fetchMore()
    assert col_model.rowCount() == 5 * ROWS_TO_LOAD


def test_editor_parent_set(monkeypatch):
    """
    Test that editors have their parent set so they close with Spyder.

    Regression test for spyder-ide/spyder#5696.
    """
    # Mocking and setup
    test_parent = QWidget()

    MockCollectionsEditor = Mock()
    attr_to_patch_coledit = ('spyder.widgets.' +
                             'collectionseditor.CollectionsEditor')
    monkeypatch.setattr(attr_to_patch_coledit, MockCollectionsEditor)

    MockArrayEditor = Mock()
    attr_to_patch_arredit = ('spyder.plugins.variableexplorer.widgets.' +
                             'arrayeditor.ArrayEditor')
    monkeypatch.setattr(attr_to_patch_arredit, MockArrayEditor)

    MockDataFrameEditor = Mock()
    attr_to_patch_dfedit = ('spyder.plugins.variableexplorer.widgets.' +
                            'dataframeeditor.DataFrameEditor')
    monkeypatch.setattr(attr_to_patch_dfedit, MockDataFrameEditor)

    MockTextEditor = Mock()
    attr_to_patch_textedit = ('spyder.plugins.variableexplorer.widgets.' +
                              'collectionsdelegate.TextEditor')
    monkeypatch.setattr(attr_to_patch_textedit, MockTextEditor)

    MockObjectExplorer = Mock()
    attr_to_patch_objectexplorer = ('spyder.plugins.variableexplorer.widgets.'
                                    + 'objectexplorer.ObjectExplorer')
    monkeypatch.setattr(attr_to_patch_objectexplorer, MockObjectExplorer)

    editor_data = [[0, 1, 2, 3, 4],
                   numpy.array([1.0, 42.0, 1337.0]),
                   pandas.DataFrame([[1, 2, 3], [20, 30, 40]]),
                   os,
                   "012345678901234567890123456789012345678901234567890123456"]
    col_editor = CollectionsEditorTableView(test_parent, editor_data)
    assert col_editor.parent() is test_parent

    for idx, mock_class in enumerate([MockCollectionsEditor,
                                      MockArrayEditor,
                                      MockDataFrameEditor,
                                      MockObjectExplorer,
                                      MockTextEditor]):
        col_editor.delegate.createEditor(col_editor.parent(), None,
                                         col_editor.model.index(idx, 3))
        assert mock_class.call_count == 1 + (idx // 4)
        assert mock_class.call_args[1]["parent"] is test_parent


def test_xml_dom_element_view():
    """
    Test that XML DOM ``Element``s are able to be viewied in CollectionsEditor.

    Regression test for spyder-ide/spyder#5642.
    """
    xml_path = path.join(LOCATION, 'dom_element_test.xml')
    with open(xml_path) as xml_file:
        xml_data = xml_file.read()

    xml_content = parseString(xml_data)
    xml_element = xml_content.getElementsByTagName("note")[0]

    col_editor = CollectionsEditor(None)
    col_editor.setup(xml_element)
    col_editor.show()
    assert col_editor.get_value()
    col_editor.accept()


def test_pandas_dateoffset_view():
    """
    Test that pandas ``DateOffset`` objs can be viewied in CollectionsEditor.

    Regression test for spyder-ide/spyder#6729.
    """
    test_dateoffset = pandas.DateOffset()
    col_editor = CollectionsEditor(None)
    col_editor.setup(test_dateoffset)
    col_editor.show()
    assert col_editor.get_value()
    col_editor.accept()


def test_set_nonsettable_objects(nonsettable_objects_data):
    """
    Test that errors trying to set attributes in ColEdit are handled properly.

    Unit regression test for issues spyder-ide/spyder#6727 and
    spyder-ide/spyder#6728.
    """
    for test_obj, expected_obj, keys in nonsettable_objects_data:
        col_model = CollectionsModel(None, test_obj)
        col_model.load_all()
        indicies = [col_model.get_index_from_key(key) for key in keys]
        for idx in indicies:
            assert not col_model.set_value(idx, "2")
            # Due to numpy's deliberate breakage of __eq__ comparison
            assert all([key == "_typ" or
                        (getattr(col_model.get_data().__obj__, key)
                         == getattr(expected_obj, key)) for key in keys])


@flaky(max_runs=3)
@pytest.mark.no_xvfb
def test_edit_nonsettable_objects(qtbot, nonsettable_objects_data):
    """
    Test that errors trying to edit attributes in ColEdit are handled properly.

    Integration regression test for issues spyder-ide/spyder#6727 and
    spyder-ide/spyder#6728.
    """
    for test_obj, expected_obj, keys in nonsettable_objects_data:
        col_editor = CollectionsEditor(None)
        col_editor.setup(test_obj)
        with qtbot.waitExposed(col_editor):
            col_editor.show()
        view = col_editor.widget.editor
        indicies = [view.source_model.get_index_from_key(key) for key in keys]

        for _ in range(3):
            qtbot.keyClick(view, Qt.Key_Right)
        last_row = -1
        rows_to_test = [index.row() for index in indicies]
        for row in rows_to_test:
            for _ in range(row - last_row - 1):
                qtbot.keyClick(view, Qt.Key_Down)
            qtbot.keyClick(view, Qt.Key_Space)
            qtbot.keyClick(view.focusWidget(), Qt.Key_Backspace)
            qtbot.keyClicks(view.focusWidget(), "2")
            qtbot.keyClick(view.focusWidget(), Qt.Key_Down)
            last_row = row

        qtbot.wait(100)

        # Due to numpy's deliberate breakage of __eq__ comparison
        assert all([key == "_typ" or (getattr(col_editor.get_value(), key)
                    == getattr(expected_obj, key)) for key in keys])

        col_editor.accept()
        qtbot.wait(200)

        # Same reason as above
        assert all([key == "_typ" or (getattr(col_editor.get_value(), key)
                    == getattr(expected_obj, key)) for key in keys])

        if getattr(test_obj, "_typ", None) is None:
            keys.remove("_typ")

        assert all([getattr(test_obj, key)
                    == getattr(expected_obj, key) for key in keys])


def test_collectionseditor_with_class_having_buggy_copy(qtbot):
    """
    Test that editor for object whose .copy() returns a different type is
    readonly; cf. spyder-ide/spyder#6936.
    """
    class MyDictWithBuggyCopy(dict):
        pass

    md = MyDictWithBuggyCopy({1: 2})
    editor = CollectionsEditor()
    editor.setup(md)
    assert editor.widget.editor.readonly


def test_collectionseditor_with_class_having_correct_copy(qtbot):
    """
    Test that editor for object whose .copy() returns the same type is not
    readonly; cf. spyder-ide/spyder#6936.
    """
    class MyDictWithCorrectCopy(dict):
        def copy(self):
            return MyDictWithCorrectCopy(self)

    md = MyDictWithCorrectCopy({1: 2})
    editor = CollectionsEditor()
    editor.setup(md)
    assert not editor.widget.editor.readonly


def test_collectionseditor_when_clicking_on_header_and_large_rows(qtbot):
    """
    Test that sorting works when clicking in its header and there's a
    large number of rows.
    """
    li = [1] * 10000
    editor = CollectionsEditor()
    editor.setup(li)
    editor.show()

    # Perform the sorting. It should be done quite quickly because
    # there's a very small number of rows in display.
    view = editor.widget.editor
    header = view.horizontalHeader()
    with qtbot.waitSignal(header.sectionClicked, timeout=200):
        qtbot.mouseClick(header.viewport(), Qt.LeftButton, pos=QPoint(1, 1))

    # Assert data was sorted correctly.
    assert data(view.model, 0, 0) == 9999

    editor.accept()


def test_dicts_with_mixed_types_as_key(qtbot):
    """
    Test that we can show dictionaries with mixed data types as keys.

    This is a regression for spyder-ide/spyder#13481.
    """
    colors = {1: 'red', 'Y': 'yellow'}
    editor = CollectionsEditor()
    editor.setup(colors)
    assert editor.widget.editor.source_model.keys == [1, 'Y']


def test_dicts_natural_sorting(qtbot):
    """
    Test that natural sorting actually does what it should do
    """
    import random
    numbers = list(range(100))
    random.shuffle(numbers)
    dictionary = {'test{}'.format(i): None for i in numbers}
    data_sorted = sorted(list(dictionary.keys()), key=natsort)
    # numbers should be as a human would sort, e.g. test3 before test100
    # regular sort would sort test1, test10, test11,..., test2, test20,...
    expected = ['test{}'.format(i) for i in list(range(100))]
    editor = CollectionsEditor()
    editor.setup(dictionary)
    editor.widget.editor.source_model.sort(0)

    assert data_sorted == expected, 'Function failed'
    assert editor.widget.editor.source_model.keys == expected, \
        'GUI sorting fail'


def test_dicts_natural_sorting_mixed_types():
    """
    Test that natural sorting actually does what it should do.
    testing for issue 13733, as mixed types were sorted incorrectly.

    Sorting for other columns will be tested as well.
    """
    import pandas as pd
    dictionary = {'DSeries': pd.Series(dtype=int), 'aStr': 'algName',
                  'kDict': {2: 'asd', 3: 2}}

    # put this here variable, as it might change later to reflect string length
    str_size = get_size(dictionary['aStr'])

    editor = CollectionsEditor()
    editor.setup(dictionary)
    cm = editor.widget.editor.source_model
    cm.sort(0)
    keys = cm.keys
    types = cm.types
    sizes = cm.sizes

    assert keys == ['aStr', 'DSeries', 'kDict']
    assert types == ['str', 'Series', 'dict']
    assert sizes == [str_size, (0,), 2]

    assert data_table(cm, 3, 3) == [['aStr', 'DSeries', 'kDict'],
                                    ['str', 'Series', 'dict'],
                                    [str_size, '(0,)', 2]]

    # insert an item and check that it is still sorted correctly
    editor.widget.editor.new_value('List', [1, 2, 3])
    assert data_table(cm, 4, 3) == [['aStr', 'DSeries', 'kDict', 'List'],
                                    ['str', 'Series', 'dict', 'list'],
                                    [str_size, '(0,)', 2, 3]]
    cm.sort(0)
    assert data_table(cm, 4, 3) == [['aStr', 'DSeries', 'kDict', 'List'],
                                    ['str', 'Series', 'dict', 'list'],
                                    [str_size, '(0,)', 2, 3]]

    # now sort for types
    cm.sort(1)
    assert data_table(cm, 4, 3) == [['DSeries', 'kDict', 'List', 'aStr'],
                                    ['Series', 'dict', 'list', 'str'],
                                    ['(0,)', 2, 3, str_size]]

    # now sort for sizes
    cm.sort(2)
    assert data_table(cm, 4, 3) == [['DSeries', 'kDict', 'List', 'aStr'],
                                    ['Series', 'dict', 'list', 'str'],
                                    ['(0,)', 2, 3, str_size]]


def test_collectioneditor_plot(qtbot):
    """
    Test that plotting a list from the collection editor calls the .plot()
    function in the associated namespace browser.
    """
    my_list = [4, 2]
    mock_namespacebrowser = Mock()
    cew = CollectionsEditorWidget(
        None, {'list': my_list}, namespacebrowser=mock_namespacebrowser)
    qtbot.addWidget(cew)

    cew.editor.plot('list', 'plot')
    mock_namespacebrowser.plot.assert_called_once_with(my_list, 'plot')


def test_collectionseditor_select_row_button(qtbot):
    """Test that the button to select rows is working as expected."""
    data = {"a": 10, "b": "This is a string"}
    editor = CollectionsEditor()
    editor.setup(data)
    editor.show()

    # This is necessary so that Qt paints
    qtbot.wait(300)

    # Coordinates to position the cursor on top of the select row button for
    # the first row
    table_view = editor.widget.editor
    x = (
        # Left x ccordinate for the first row
        + table_view.columnViewportPosition(0)
        + table_view.width()
        - SELECT_ROW_BUTTON_SIZE // 2
    )

    y = (
        # Top y ccordinate for the first row
        + table_view.rowViewportPosition(0)
        + table_view.rowHeight(0) // 2
    )

    # Move cursor
    qtbot.mouseMove(table_view.viewport(), QPoint(x, y), delay=100)

    # Click on that posiiton and check the first row was selected.
    # Note: We can't use LeftButton here because it edits the row. However, it
    # works as exoected in regular usage.
    qtbot.mouseClick(table_view.viewport(), Qt.MiddleButton, pos=QPoint(x, y))
    assert table_view.selected_rows() == {0}

    # Click again and check the row was deselected
    qtbot.mouseClick(table_view.viewport(), Qt.MiddleButton, pos=QPoint(x, y))
    assert table_view.selected_rows() == set()


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from __future__ import annotations
import asyncio
import json
import logging
import socket

import asyncssh
from packaging.version import Version

from spyder.api.translations import _
from spyder.config.base import get_debug_level
from spyder.plugins.remoteclient import (
    SPYDER_REMOTE_MAX_VERSION,
    SPYDER_REMOTE_MIN_VERSION,
)
from spyder.plugins.remoteclient.api.jupyterhub import JupyterAPI
from spyder.plugins.remoteclient.api.protocol import (
    ConnectionInfo,
    ConnectionStatus,
    KernelConnectionInfo,
    KernelInfo,
    RemoteClientLog,
    SSHClientOptions,
)
from spyder.plugins.remoteclient.api.ssh import SpyderSSHClient
from spyder.plugins.remoteclient.utils.installation import (
    get_installer_command,
    get_server_version_command,
    SERVER_ENV,
)


class SpyderRemoteClientLoggerHandler(logging.Handler):
    def __init__(self, client, *args, **kwargs):
        self._client = client
        super().__init__(*args, **kwargs)

        log_format = "%(message)s &#8212; %(asctime)s"
        formatter = logging.Formatter(
            log_format, datefmt="%H:%M:%S %d/%m/%Y"
        )
        self.setFormatter(formatter)

    def emit(self, record):
        self._client._plugin.sig_client_message_logged.emit(
            RemoteClientLog(
                id=self._client.config_id,
                message=self.format(record),
                level=record.levelno,
                created=record.created,
            )
        )


class SpyderRemoteClient:
    """Class to manage a remote server and its kernels."""

    JUPYTER_SERVER_TIMEOUT = 5  # seconds

    _extra_options = ["platform", "id"]

    START_SERVER_COMMAND = f"/${{HOME}}/.local/bin/micromamba run -n {SERVER_ENV} spyder-server --jupyter-server"
    CHECK_SERVER_COMMAND = f"/${{HOME}}/.local/bin/micromamba run -n {SERVER_ENV} spyder-server -h"
    GET_SERVER_INFO_COMMAND = f"/${{HOME}}/.local/bin/micromamba run -n {SERVER_ENV} spyder-server --get-running-info"

    def __init__(self, conf_id, options: SSHClientOptions, _plugin=None):
        self._config_id = conf_id
        self.options = options
        self._plugin = _plugin

        self.__server_installed = asyncio.Event()
        self.__server_started = asyncio.Event()
        self.__connection_established = asyncio.Event()
        self.__installing_server = False
        self.__starting_server = False
        self.__creating_connection = False

        self._ssh_connection: asyncssh.SSHClientConnection = None
        self._remote_server_process: asyncssh.SSHClientProcess = None
        self._port_forwarder: asyncssh.SSHListener = None
        self._server_info = {}

        # For logging
        self._logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}({self.config_id})"
        )

        if not get_debug_level():
            self._logger.setLevel(logging.DEBUG)

        if self._plugin is not None:
            self._logger.addHandler(SpyderRemoteClientLoggerHandler(self))

    def __emit_connection_status(self, status, message):
        if self._plugin is not None:
            self._plugin.sig_connection_status_changed.emit(
                ConnectionInfo(
                    id=self.config_id, status=status, message=message
                )
            )

    def __emit_version_mismatch(self, version: str):
        if self._plugin is not None:
            self._plugin.sig_version_mismatch.emit(self.config_id, version)

    @property
    def _api_token(self):
        return self._server_info.get("token")

    @property
    def server_port(self):
        return self._server_info.get("port")

    @property
    def server_pid(self):
        return self._server_info.get("pid")

    @property
    def server_started(self):
        return self.__server_started.is_set() and not self.__starting_server

    @property
    def ssh_is_connected(self):
        """Check if SSH connection is open."""
        return (
            self.__connection_established.is_set()
            and not self.__creating_connection
        )

    @property
    def port_is_forwarded(self):
        """Check if local port is forwarded."""
        return self._port_forwarder is not None

    @property
    def config_id(self):
        """Return the configuration ID"""
        return self._config_id

    @property
    def server_url(self):
        """Return server URL

        Get the server URL with the fowarded local port that
        is used to connect to the spyder-remote-server.

        Returns
        -------
        str
            JupyterHub server URL

        Raises
        ------
        ValueError
            If the local port is not set.
        """
        if not self.local_port:
            raise ValueError("Local port is not set")
        return f"http://127.0.0.1:{self.local_port}"

    @property
    def api_token(self):
        if not self._api_token:
            raise ValueError("API token is not set")
        return self._api_token

    @property
    def peer_host(self):
        if self._ssh_connection is not None:
            return self._ssh_connection.get_extra_info("peername")[0]
        else:
            return None

    @property
    def peer_port(self):
        if not self.ssh_is_connected:
            return None

        return self._ssh_connection.get_extra_info("peername")[1]

    @property
    def peer_username(self):
        if not self.ssh_is_connected:
            return None

        return self._ssh_connection.get_extra_info("username")

    @property
    def client_factory(self):
        """Return the client factory."""
        return lambda: SpyderSSHClient(self)

    async def close(self):
        """Closes the remote server and the SSH connection."""
        self.__emit_connection_status(
            ConnectionStatus.Stopping,
            _("We're closing the connection. Please be patient"),
        )

        await self.stop_remote_server()
        await self.close_port_forwarder()
        await self.close_ssh_connection()

    def _handle_connection_lost(self, exc: Exception | None = None):
        self.__connection_established.clear()
        self.__server_started.clear()
        self._port_forwarder = None
        if exc:
            self._logger.error(
                f"Connection to {self.peer_host} was lost",
                exc_info=exc,
            )
            self.__emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("The connection was lost"),
            )

    async def get_server_info(self):
        """Check if the remote server is running."""
        if self._ssh_connection is None:
            self._logger.debug("ssh connection was not established")
            return None

        try:
            output = await self._ssh_connection.run(
                self.GET_SERVER_INFO_COMMAND, check=True
            )
        except asyncssh.TimeoutError:
            self._logger.error("Getting server info timed out")
            return None
        except asyncssh.misc.ChannelOpenError:
            self._logger.error(
                "The connection is closed, so it's not possible to get the "
                "server info"
            )
            return None
        except asyncssh.ProcessError as err:
            self._logger.debug(f"Error getting server info: {err.stderr}")
            return None

        try:
            info = json.loads(output.stdout.splitlines()[-1])
        except json.JSONDecodeError:
            self._logger.debug(
                f"Issue parsing server info: {output.stdout}"
            )
            return None

        return info

    # -- Connection and server management
    async def connect_and_install_remote_server(self) -> bool:
        """Connect to the remote server and install the server."""
        if await self.create_new_connection():
            return await self.install_remote_server()

        return False

    async def connect_and_start_server(self) -> bool:
        """Connect to the remote server and start the server."""
        if await self.create_new_connection():
            return await self.start_remote_server()

        return False

    async def ensure_connection_and_server(self) -> bool:
        """
        Ensure the SSH connection is open and the remote server is running.
        """
        if self.ssh_is_connected and not self.server_started:
            return await self.ensure_server()

        if not self.ssh_is_connected:
            return await self.connect_and_ensure_server()

        return True

    async def connect_and_ensure_server(self) -> bool:
        """
        Connect to the remote server and ensure it is installed and running.
        """
        if await self.create_new_connection() and not self.server_started:
            return await self.ensure_server()

        if self.server_started:
            return True

        return False

    async def ensure_connection(self) -> bool:
        """Ensure the SSH connection is open."""
        if self.ssh_is_connected:
            return True

        return await self.create_new_connection()

    async def ensure_server(self, *, check_installed=True) -> bool:
        """Ensure remote server is installed and running."""
        if self.server_started:
            return True

        if check_installed and not await self.ensure_server_installed():
            return False

        return await self.start_remote_server()

    async def start_remote_server(self):
        """Start remote server."""
        if self.__starting_server:
            await self.__server_started.wait()
            return True

        self.__starting_server = True
        try:
            if await self.__start_remote_server():
                self.__server_started.set()
                return True
        finally:
            self.__starting_server = False

        self.__server_started.clear()
        return False

    async def __start_remote_server(self):
        """Start remote server."""
        if not self.ssh_is_connected:
            self._logger.error("SSH connection is not open")
            self.__emit_connection_status(
                ConnectionStatus.Error,
                _("The SSH connection is not open"),
            )
            return False

        if info := await self.get_server_info():
            self._logger.warning(
                f"Remote server is already running for {self.peer_host}"
            )

            self._logger.debug("Checking server info")
            if self._server_info != info:
                self._server_info = info
                self._logger.info(
                    "Different server info, updating info "
                    f"for {self.peer_host}"
                )
                if await self.forward_local_port():
                    self.__emit_connection_status(
                        ConnectionStatus.Active,
                        _("The connection was established successfully"),
                    )
                    return True

                self._logger.error(
                    "Error forwarding local port, server might not be "
                    "reachable"
                )
                self.__emit_connection_status(
                    ConnectionStatus.Error,
                    _("It was not possible to forward the local port"),
                )

            self.__emit_connection_status(
                ConnectionStatus.Active,
                _("The connection was established successfully"),
            )

            return True

        self._logger.debug(f"Starting remote server for {self.peer_host}")
        try:
            self._remote_server_process = (
                await self._ssh_connection.create_process(
                    self.START_SERVER_COMMAND,
                    stderr=asyncssh.STDOUT,
                )
            )
        except (OSError, asyncssh.Error, ValueError) as e:
            self._logger.error(f"Error starting remote server: {e}")
            self._remote_server_process = None
            self.__emit_connection_status(
                ConnectionStatus.Error, _("Error starting the remote server")
            )
            return False

        _time = 0
        while (info := await self.get_server_info()) is None and _time < 5:
            await asyncio.sleep(1)
            _time += 1

        if info is None:
            self._logger.error("Faield to get server info")
            self.__emit_connection_status(
                ConnectionStatus.Error,
                _(
                    "There was an error when trying to get the remote server "
                    "information"
                ),
            )
            return False

        self._server_info = info

        self._logger.info(
            f"Remote server started for {self.peer_host} at port "
            f"{self.server_port}"
        )

        if await self.forward_local_port():
            self.__emit_connection_status(
                ConnectionStatus.Active,
                _("The connection was established successfully"),
            )
            return True

        self._logger.error("Error forwarding local port.")
        self.__emit_connection_status(
            ConnectionStatus.Error,
            _("It was not possible to forward the local port"),
        )
        return False

    async def ensure_server_installed(self) -> bool:
        """Check remote server version."""
        if not self.ssh_is_connected:
            self._logger.error("SSH connection is not open")
            self.__emit_connection_status(
                ConnectionStatus.Error,
                _("The SSH connection is not open"),
            )
            return ""

        commnad = get_server_version_command(self.options["platform"])

        try:
            output = await self._ssh_connection.run(
                commnad, check=True
            )
        except asyncssh.ProcessError as err:
            # Server is not installed
            self._logger.warning(
                f"Issue checking server version: {err.stderr}"
            )
            return await self.install_remote_server()
        except asyncssh.TimeoutError:
            self._logger.error("Checking server version timed out")
            self.__emit_connection_status(
                ConnectionStatus.Error,
                _("The server version check timed out"),
            )
            return False

        version = output.stdout.splitlines()[-1].strip()

        if Version(version) >= Version(SPYDER_REMOTE_MAX_VERSION):
            self._logger.error(
                f"Server version mismatch: {version} is greater than "
                f"the maximum supported version {SPYDER_REMOTE_MAX_VERSION}"
            )
            self.__emit_version_mismatch(version)
            self.__emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("Error connecting to the remote server"),
            )
            return False

        if Version(version) < Version(SPYDER_REMOTE_MIN_VERSION):
            self._logger.warning(
                f"Server version mismatch: {version} is lower than "
                f"the minimum supported version {SPYDER_REMOTE_MIN_VERSION}. "
                f"A more recent version will be installed."
            )
            return await self.install_remote_server()

        self._logger.info(f"Supported Server version: {version}")

        return True

    async def install_remote_server(self) -> bool:
        """Install remote server."""
        if self.__installing_server:
            await self.__server_installed.wait()
            return True

        self.__installing_server = True
        try:
            if await self.__install_remote_server():
                self.__server_installed.set()
                return True
        finally:
            self.__installing_server = False

        self.__server_installed.clear()
        return False

    async def __install_remote_server(self):
        """Install remote server."""
        if not self.ssh_is_connected:
            self._logger.error("SSH connection is not open")
            self.__emit_connection_status(
                ConnectionStatus.Error,
                _("The SSH connection is not open"),
            )
            return False

        self._logger.debug(
            f"Installing spyder-remote-server on {self.peer_host}"
        )

        try:
            command = get_installer_command(self.options["platform"])
        except NotImplementedError:
            self._logger.error(
                f"Cannot install spyder-remote-server on "
                f"{self.options['platform']} automatically. Please install it "
                f"manually."
            )
            self.__emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("There was an error installing the remote server"),
            )
            return False

        try:
            await self._ssh_connection.run(command, check=True)
        except asyncssh.ProcessError as err:
            self._logger.error(f"Installation script failed: {err.stderr}")
            self.__emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("There was an error installing the remote server"),
            )
            return False
        except asyncssh.TimeoutError:
            self._logger.error("Installation script timed out")
            self.__emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("There was an error installing the remote server"),
            )
            return False

        self._logger.info(
            f"Successfully installed spyder-remote-server on {self.peer_host}"
        )

        return True

    async def create_new_connection(self) -> bool:
        if self.__creating_connection:
            await self.__connection_established.wait()
            return True

        self.__creating_connection = True
        try:
            if await self.__create_new_connection():
                self.__connection_established.set()
                return True
        finally:
            self.__creating_connection = False

        self.__connection_established.clear()
        return False

    async def __create_new_connection(self) -> bool:
        """Creates a new SSH connection to the remote server machine.

        Args
        ----
        options: dict[str, str]
            The options to use for the SSH connection.

        Returns
        -------
        bool
            True if the connection was successful, False otherwise.
        """
        if self.ssh_is_connected:
            self._logger.debug(
                f"Atempting to create a new connection with an existing for "
                f"{self.peer_host}"
            )
            await self.close_ssh_connection()

        self.__emit_connection_status(
            ConnectionStatus.Connecting,
            _("We're establishing the connection. Please be patient"),
        )

        connect_kwargs = {
            k: v
            for k, v in self.options.items()
            if k not in self._extra_options
        }
        self._logger.debug("Opening SSH connection")
        try:
            self._ssh_connection = await asyncssh.connect(
                **connect_kwargs, client_factory=self.client_factory
            )
        except (OSError, asyncssh.Error) as e:
            self._logger.error(f"Failed to open ssh connection: {e}")
            self.__emit_connection_status(
                ConnectionStatus.Error,
                _("It was not possible to open a connection to this machine"),
            )
            return False

        self._logger.info(f"SSH connection opened for {self.peer_host}")

        return True

    async def forward_local_port(self):
        """Forward local port."""
        if not self.server_port:
            self._logger.error("Server port is not set")
            self.__emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("The server port is not set"),
            )
            return False

        if not self.ssh_is_connected:
            self._logger.error("SSH connection is not open")
            self.__emit_connection_status(
                status=ConnectionStatus.Error,
                message=_("The SSH connection is not open"),
            )
            return False

        self._logger.debug(
            f"Forwarding an free local port to remote port {self.server_port}"
        )

        if self._port_forwarder:
            self._logger.warning(
                f"Port forwarder is already open for host {self.peer_host} "
                f"with local port {self.local_port} and remote port "
                f"{self.server_port}"
            )
            await self.close_port_forwarder()

        local_port = self.get_free_port()

        server_host = self._server_info["hostname"]

        self._port_forwarder = await self._ssh_connection.forward_local_port(
            "",
            local_port,
            server_host,
            self.server_port,
        )

        self.local_port = local_port

        self._logger.debug(
            f"Forwarded local port {local_port} to remote server at "
            f"{server_host}:{self.server_port}"
        )

        return True

    async def close_port_forwarder(self):
        """Close port forwarder."""
        if self.port_is_forwarded:
            self._logger.debug(
                f"Closing port forwarder for host {self.peer_host} with local "
                f"port {self.local_port}"
            )
            self._port_forwarder.close()
            await self._port_forwarder.wait_closed()
            self._port_forwarder = None
            self._logger.debug(
                f"Port forwarder closed for host {self.peer_host} with local "
                f"port {self.local_port}"
            )

    async def stop_remote_server(self):
        """Close remote server."""
        if not self.server_started:
            self._logger.warning(
                f"Remote server is not running for {self.peer_host}"
            )
            return False

        if not self.ssh_is_connected:
            self._logger.error("SSH connection is not open")
            self.__emit_connection_status(
                ConnectionStatus.Error,
                _("The SSH connection is not open"),
            )
            return False

        # bug in jupyterhub, need to send SIGINT twice
        self._logger.debug(
            f"Stopping remote server for {self.peer_host} with pid "
            f"{self._server_info['pid']}"
        )
        try:
            async with JupyterAPI(
                self.server_url, api_token=self.api_token
            ) as jupyter:
                await jupyter.shutdown_server()
        except Exception as err:
            self._logger.exception(
                "Error stopping remote server", exc_info=err
            )

        if (
            self._remote_server_process
            and not self._remote_server_process.is_closing()
        ):
            self._remote_server_process.terminate()
            await self._remote_server_process.wait_closed()

        self.__server_started.clear()
        self._remote_server_process = None
        self._logger.info(f"Remote server process closed for {self.peer_host}")
        return True

    async def close_ssh_connection(self):
        """Close SSH connection."""
        if not self.ssh_is_connected:
            self._logger.debug("SSH connection is not open")
            return

        self._logger.debug(f"Closing SSH connection for {self.peer_host}")
        self._ssh_connection.close()
        await self._ssh_connection.wait_closed()
        self._ssh_connection = None
        self._logger.info("SSH connection closed")
        self.__connection_established.clear()
        self.__emit_connection_status(
            ConnectionStatus.Inactive,
            _("The connection was closed successfully"),
        )

    # --- Kernel Management
    async def start_new_kernel_ensure_server(
        self, _retries=5
    ) -> KernelConnectionInfo:
        """Launch a new kernel ensuring the remote server is running.

        Parameters
        ----------
        options : SSHClientOptions
            The options to use for the SSH connection.

        Returns
        -------
        KernelConnectionInfo
            The kernel connection information.
        """
        if not await self.ensure_connection_and_server():
            self._logger.error(
                "Cannot launch kernel, remote server is not running"
            )
            return {}

        # This is necessary to avoid an error when the server has not started
        # before
        await asyncio.sleep(1)
        kernel_id = await self.start_new_kernel()

        retries = 0
        while not kernel_id and retries < _retries:
            await asyncio.sleep(1)
            kernel_id = await self.start_new_kernel()
            self._logger.debug(
                f"Server might not be ready yet, retrying kernel launch "
                f"({retries + 1}/{_retries})"
            )
            retries += 1

        return kernel_id

    async def get_kernel_info_ensure_server(
        self, kernel_id, _retries=5
    ) -> KernelConnectionInfo:
        """Launch a new kernel ensuring the remote server is running.

        Parameters
        ----------
        options : SSHClientOptions
            The options to use for the SSH connection.

        Returns
        -------
        KernelConnectionInfo
            The kernel connection information.
        """
        if not await self.ensure_connection_and_server():
            self._logger.error(
                "Cannot launch kernel, remote server is not running"
            )
            return {}

        # This is necessary to avoid an error when the server has not started
        # before
        await asyncio.sleep(1)
        kernel_info = await self.get_kernel_info(kernel_id)

        retries = 0
        while not kernel_info and retries < _retries:
            await asyncio.sleep(1)
            kernel_info = await self.get_kernel_info(kernel_id)
            self._logger.debug(
                f"Server might not be ready yet, retrying kernel launch "
                f"({retries + 1}/{_retries})"
            )
            retries += 1

        return kernel_info

    async def start_new_kernel(self, kernel_spec=None) -> KernelInfo:
        """Start new kernel."""
        async with JupyterAPI(
            self.server_url, api_token=self.api_token
        ) as jupyter:
            response = await jupyter.create_kernel(kernel_spec=kernel_spec)
        self._logger.info(f"Kernel started with ID {response['id']}")
        return response

    async def list_kernels(self) -> list[KernelInfo]:
        """List kernels."""
        async with JupyterAPI(
            self.server_url, api_token=self.api_token
        ) as jupyter:
            response = await jupyter.list_kernels()

        self._logger.info(f"Kernels listed for {self.peer_host}")
        return response

    async def get_kernel_info(self, kernel_id) -> KernelInfo:
        """Get kernel info."""
        async with JupyterAPI(
            self.server_url, api_token=self.api_token
        ) as jupyter:
            response = await jupyter.get_kernel(kernel_id=kernel_id)

        self._logger.info(f"Kernel info retrieved for ID {kernel_id}")
        return response

    async def terminate_kernel(self, kernel_id) -> bool:
        """Terminate kernel."""
        async with JupyterAPI(
            self.server_url, api_token=self.api_token
        ) as jupyter:
            response = await jupyter.delete_kernel(kernel_id=kernel_id)

        self._logger.info(f"Kernel terminated for ID {kernel_id}")
        return response

    async def interrupt_kernel(self, kernel_id) -> bool:
        """Interrupt kernel."""
        async with JupyterAPI(
            self.server_url, api_token=self.api_token
        ) as jupyter:
            response = await jupyter.interrupt_kernel(kernel_id=kernel_id)

        self._logger.info(f"Kernel interrupted for ID {kernel_id}")
        return response

    async def restart_kernel(self, kernel_id) -> bool:
        """Restart kernel."""
        async with JupyterAPI(
            self.server_url, api_token=self.api_token
        ) as jupyter:
            response = await jupyter.restart_kernel(kernel_id=kernel_id)

        self._logger.info(f"Kernel restarted for ID {kernel_id}")
        return response

    @staticmethod
    def get_free_port():
        """Request a free port from the OS."""
        with socket.socket() as s:
            s.bind(("", 0))
            return s.getsockname()[1]



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Tests for the Editor plugin."""

# Standard library imports
import os
import os.path as osp
import shutil

# Third party imports
from qtpy.QtCore import Qt
import pytest

# Local imports
from spyder.api.plugins import Plugins
from spyder.plugins.editor.utils.autosave import AutosaveForPlugin
from spyder.plugins.editor.widgets.editorstack import editorstack as editor_module
from spyder.plugins.editor.widgets.codeeditor import CodeEditor
from spyder.plugins.run.api import RunContext
from spyder.utils.sourcecode import get_eol_chars, get_eol_chars_from_os_name


# =============================================================================
# ---- Tests
# =============================================================================
def test_basic_initialization(editor_plugin):
    """Test Editor plugin initialization."""
    editor = editor_plugin

    # Assert that editor exists
    assert editor is not None


@pytest.mark.parametrize(
    'last_focused_filename, expected_current_filename',
    [('other_file.py', 'file1.py'),
     ('file1.py', 'file1.py'),
     ('file2.py', 'file2.py'),
     ('file4.py', 'file4.py')
     ])
def test_setup_open_files(editor_plugin_open_files, last_focused_filename,
                          expected_current_filename):
    """Test Editor plugin open files setup.

    Test that the file order is preserved during the Editor plugin setup and
    that the current file correspond to the last focused file.
    """
    editor_factory = editor_plugin_open_files
    editor, expected_filenames, expected_current_filename = (
        editor_factory(last_focused_filename, expected_current_filename))

    current_filename = editor.get_current_editorstack().get_current_filename()
    current_filename = osp.normcase(current_filename)
    assert current_filename == expected_current_filename
    filenames = editor.get_current_editorstack().get_filenames()
    filenames = [osp.normcase(f) for f in filenames]
    assert filenames == expected_filenames


def test_restore_open_files(qtbot, editor_plugin_open_files):
    """Test restoring of opened files without Projects plugin"""
    editor_factory = editor_plugin_open_files
    editor, expected_filenames, expected_current_filename = (
        editor_factory(None, None))

    # Pre-condition: Projects plugin is disabled
    assert editor.get_plugin(Plugins.Projects, error=False) is None

    # `expected_filenames` is modified. A copy is required because
    # `expected_filenames` and `editor.get_conf("filesnames")` are the same
    # object.
    expected_filenames = expected_filenames.copy()
    assert expected_filenames is not editor.get_conf("filenames")
    for i in range(2):
        filename = expected_filenames.pop()
        editor.close_file_from_name(filename)

    # Close editor and check that opened files are saved
    editor.on_close()
    filenames = [osp.normcase(f) for f in editor.get_conf("filenames")]
    assert filenames == expected_filenames

    # “Re-open” editor and check the opened files are restored
    editor.setup_open_files(close_previous_files=True)
    filenames = editor.get_current_editorstack().get_filenames()
    filenames = [osp.normcase(f) for f in filenames]
    assert filenames == expected_filenames


def test_setup_open_files_cleanprefs(editor_plugin_open_files):
    """Test that Editor successfully opens files if layout is not defined.

    Regression test for spyder-ide/spyder#8458.
    """
    editor_factory = editor_plugin_open_files
    editor, expected_filenames, expected_current_filename = (
        editor_factory(None, None))

    filenames = editor.get_current_editorstack().get_filenames()
    filenames = [osp.normcase(f) for f in filenames]
    assert filenames == expected_filenames
    current_filename = editor.get_current_editorstack().get_current_filename()
    current_filename = osp.normcase(current_filename)
    assert current_filename == expected_current_filename


def test_open_untitled_files(editor_plugin_open_files):
    """
    Test for checking the counter of the untitled files is starting
    correctly when there is one or more `untitledx.py` files saved.

    Regression test for spyder-ide/spyder#7831
    """
    editor_factory = editor_plugin_open_files
    editor, expected_filenames, expected_current_filename = (
        editor_factory(None, None))

    editor.new()
    filenames = editor.get_current_editorstack().get_filenames()
    new_filename = filenames[-1]
    assert 'untitled5.py' in new_filename


def test_renamed_tree(editor_plugin, mocker):
    """Test editor.renamed_tree().

    This tests that the file renaming functions are called correctly,
    but does not test that all the renaming happens in File Explorer,
    Project Explorer, and Editor widget as those aren't part of the plugin.
    """
    editor = editor_plugin
    editor_main_widget = editor.get_widget()
    mocker.patch.object(editor_main_widget, 'get_filenames')
    mocker.patch.object(editor_main_widget, 'renamed')
    if os.name == "nt":
        filenames = [r'C:\test\directory\file1.py',
                     r'C:\test\directory\file2.txt',
                     r'C:\home\spyder\testing\file3.py',
                     r'C:\test\directory\file4.rst']
        expected = [r'C:\test\dir\file1.py',
                    r'C:\test\dir\file2.txt',
                    r'C:\home\spyder\testing\file3.py',
                    r'C:\test\dir\file4.rst']
        sourcedir = r'C:\test\directory'
        destdir = r'C:\test\dir'
    else:
        filenames = ['/test/directory/file1.py',
                     '/test/directory/file2.txt',
                     '/home/spyder/testing/file3.py',
                     '/test/directory/file4.rst']
        expected = ['/test/dir/file1.py',
                    '/test/dir/file2.txt',
                    '/home/spyder/testing/file3.py',
                    '/test/dir/file4.rst']
        sourcedir = '/test/directory'
        destdir = '/test/dir'

    editor_main_widget.get_filenames.return_value = filenames

    editor_main_widget.renamed_tree(sourcedir, destdir)
    assert editor_main_widget.renamed.call_count == 3
    for file in [0, 1, 3]:
        editor_main_widget.renamed.assert_any_call(
            source=filenames[file],
            dest=expected[file]
        )


def test_no_template(editor_plugin):
    """
    Test that new files can be opened when no template is found.
    """
    editor = editor_plugin

    # Move template to another file to simulate the lack of it
    template = editor.get_widget().TEMPLATE_PATH
    shutil.move(template, osp.join(osp.dirname(template), 'template.py.old'))

    # Open a new file
    editor.new()

    # Get contents
    code_editor = editor.get_widget().get_focus_widget()
    contents = code_editor.get_text('sof', 'eof')

    # Assert contents are empty
    assert not contents

    # Revert template back
    shutil.move(osp.join(osp.dirname(template), 'template.py.old'), template)


def test_editor_has_autosave_component(editor_plugin):
    """Test that Editor includes an AutosaveForPlugin."""
    editor = editor_plugin
    assert isinstance(editor.get_widget().autosave, AutosaveForPlugin)


def test_autosave_component_do_autosave(editor_plugin, mocker):
    """Test that AutosaveForPlugin's do_autosave() calls the current editor
    stack's autosave_all()."""
    editor = editor_plugin
    editorStack = editor.get_current_editorstack()
    mocker.patch.object(editorStack.autosave, 'autosave_all')
    editor.get_widget().autosave.do_autosave()
    assert editorStack.autosave.autosave_all.called


def test_editorstacks_share_autosave_data(editor_plugin, qtbot):
    """Check that two EditorStacks share the same autosave data."""
    editor = editor_plugin.get_widget()
    editor.editorsplitter.split()
    assert len(editor.editorstacks) == 2
    autosave1 = editor.editorstacks[0].autosave
    autosave2 = editor.editorstacks[1].autosave
    assert autosave1.name_mapping is autosave2.name_mapping
    assert autosave1.file_hashes is autosave2.file_hashes


# The mock_RecoveryDialog fixture needs to be called before setup_editor, so
# it needs to be mentioned first
def test_editor_calls_recoverydialog_exec_if_nonempty(
    mock_RecoveryDialog, editor_plugin
):
    """
    Check that the editor tries to exec a recovery dialog before the main
    window is visible.
    """
    editor_plugin.before_mainwindow_visible()
    assert mock_RecoveryDialog.return_value.exec_if_nonempty.called


def test_closing_editor_plugin_stops_autosave_timer(editor_plugin):
    editor = editor_plugin
    assert editor.get_widget().autosave.timer.isActive()
    editor.get_widget().close()
    assert not editor.get_widget().autosave.timer.isActive()


def test_renamed_propagates_to_autosave(editor_plugin_open_files, mocker):
    """Test that editor.renamed() propagates info to autosave component if,
    and only if, renamed file is open in editor.

    Regression test for spyder-ide/spyder#11348"""
    editor_factory = editor_plugin_open_files
    editor, expected_filenames, expected_current_filename = (
        editor_factory(None, None))

    editorstack = editor.get_current_editorstack()
    mocker.patch.object(editorstack, 'rename_in_data')
    mocker.patch.object(editorstack.autosave, 'file_renamed')

    # Test renaming a file that is not opened in the editor
    editor.renamed('nonexisting', 'newname')
    assert not editorstack.autosave.file_renamed.called

    # Test renaming a file that is opened in the editor
    filename = editorstack.get_filenames()[0]
    editor.renamed(filename, 'newname')
    assert editorstack.autosave.file_renamed.called


def test_go_to_prev_next_cursor_position(editor_plugin, python_files):
    """
    Test the previous and next cursor position feature of the Editor.

    Regression test for spyder-ide/spyder#8000.
    """
    filenames, tmpdir = python_files
    main_widget = editor_plugin.get_widget()
    editorstack = editor_plugin.get_current_editorstack()

    expected_cursor_undo_history = []
    assert main_widget.cursor_undo_history == expected_cursor_undo_history
    # Load the Python test files (4).
    editor_plugin.load(filenames)
    # Open a new file.
    editor_plugin.new()
    # Go to the third file.
    editorstack.set_stack_index(2)
    # Move the cursor within the third file. Note that this new position is
    # not added to the cursor position history.
    editorstack.get_current_editor().set_cursor_position(5)
    # Note that we use the filenames from the editor to define the expected
    # results because those returned by the python_files fixture are
    # normalized, so this would cause issues when assessing the results.
    filenames = editor_plugin.get_filenames()
    expected_cursor_undo_history = [
        (filenames[0], 0),
        (filenames[-1], len(editorstack.data[-1].get_source_code())),
        (filenames[2], 5)
        ]
    for history, expected_history in zip(main_widget.cursor_undo_history,
                                         expected_cursor_undo_history):
        assert history[0] == expected_history[0]
        assert history[1].position() == expected_history[1]

    # Navigate to previous and next cursor positions.

    # The last entry in the cursor position history is overridden by the
    # current cursor position when going to previous or next cursor position,
    # so we need to update the last item of the expected_cursor_undo_history.
    expected_cursor_undo_history[-1] = (filenames[2], 5)

    cursor_index_moves = [-1, 1, 1, -1, -1, -1, 1, -1]
    expected_cursor_pos_indexes = [1, 2, 2, 1, 0, 0, 1, 0]
    for move, index in zip(cursor_index_moves, expected_cursor_pos_indexes):
        if move == -1:
            main_widget.go_to_previous_cursor_position()
        elif move == 1:
            main_widget.go_to_next_cursor_position()
        assert len(main_widget.cursor_undo_history) - 1 == index
        assert (editor_plugin.get_current_filename(),
                editor_plugin.get_current_editor().get_position('cursor')
                ) == expected_cursor_undo_history[index]

    for history, expected_history in zip(main_widget.cursor_undo_history,
                                         expected_cursor_undo_history[:1]):
        assert history[0] == expected_history[0]
        assert history[1].position() == expected_history[1]
    for history, expected_history in zip(main_widget.cursor_redo_history,
                                         expected_cursor_undo_history[:0:-1]):
        assert history[0] == expected_history[0]
        assert history[1].position() == expected_history[1]

    # So we are now expected to be at index 0 in the cursor position history.
    # From there, we go to the fourth file.
    editorstack.set_stack_index(3)

    # We expect that our last action caused the cursor position history to
    # be stripped from the current cursor position index and that the
    # new cursor position is added at the end of the cursor position history.
    expected_cursor_undo_history = expected_cursor_undo_history[:1]
    expected_cursor_undo_history.append((filenames[3], 0))

    for history, expected_history in zip(main_widget.cursor_undo_history,
                                         expected_cursor_undo_history):
        assert history[0] == expected_history[0]
        assert history[1].position() == expected_history[1]
    assert main_widget.cursor_redo_history == []


def test_open_and_close_lsp_requests(editor_plugin_open_files, mocker):
    """
    Test that we send the right LSP requests when opening and closing
    files.
    """
    # Patch methods whose calls we want to check
    mocker.patch.object(CodeEditor, "document_did_open")
    mocker.patch.object(CodeEditor, "notify_close")

    # Create files
    editor_factory = editor_plugin_open_files
    editor, expected_filenames, expected_current_filename = (
        editor_factory(None, None))

    # Assert that we called document_did_open once per file
    assert CodeEditor.document_did_open.call_count == 5

    # Generate a vertical split
    editorstack = editor.get_current_editorstack()
    editorstack.sig_split_vertically.emit()

    # Assert the current codeeditor has is_cloned as True
    codeeditor = editor.get_current_editor()
    assert codeeditor.is_cloned

    # Assert the number of calls to document_did_open is exactly the
    # same as before.
    assert CodeEditor.document_did_open.call_count == 5

    # Close cloned editor to verify that notify_close is called from it.
    assert CodeEditor.notify_close.call_count == 0
    editorstack = editor.get_current_editorstack()
    editorstack.close_file()
    assert CodeEditor.notify_close.call_count == 2

    # Assert focus is left in the cloned editorstack
    assert editorstack.get_current_editor().is_cloned

    # Close cloned editorstack to verify that notify_close is not called
    editorstack.close_split()
    assert CodeEditor.notify_close.call_count == 2


@pytest.mark.parametrize('os_name', ['nt', 'mac', 'posix'])
def test_toggle_eol_chars(editor_plugin, python_files, qtbot, os_name):
    """
    Check that changing eol chars from the 'Convert end-of-line characters'
    menu works as expected.
    """
    filenames, tmpdir = python_files
    editorstack = editor_plugin.get_current_editorstack()

    # Load a test file
    fname = filenames[0]
    editor_plugin.load(fname)
    qtbot.wait(500)
    codeeditor = editor_plugin.get_current_editor()

    # Change to a different eol, save and check that file has the right eol.
    editor_plugin.get_widget().toggle_eol_chars(os_name, True)
    assert codeeditor.document().isModified()
    editorstack.save()
    with open(fname, mode='r', newline='') as f:
        text = f.read()
    assert get_eol_chars(text) == get_eol_chars_from_os_name(os_name)


@pytest.mark.order(1)
@pytest.mark.parametrize('os_name', ['nt', 'mac', 'posix'])
def test_save_with_preferred_eol_chars(editor_plugin, python_files, qtbot,
                                       os_name):
    """Check that saving files with preferred eol chars works as expected."""
    filenames, tmpdir = python_files
    editorstack = editor_plugin.get_current_editorstack()
    eol_lookup = {'posix': 'LF', 'nt': 'CRLF', 'mac': 'CR'}

    # Check default options value are set
    qtbot.waitUntil(
        lambda:
            not editorstack.convert_eol_on_save
            and editorstack.convert_eol_on_save_to == 'LF'
    )


    # Load a test file
    fname = filenames[0]
    editor_plugin.load(fname)
    qtbot.wait(500)
    codeeditor = editor_plugin.get_current_editor()

    # Set options
    editor_plugin.set_conf('convert_eol_on_save', True)
    editor_plugin.set_conf('convert_eol_on_save_to', eol_lookup[os_name])
    qtbot.waitUntil(
        lambda:
            editorstack.convert_eol_on_save
            and editorstack.convert_eol_on_save_to == eol_lookup[os_name]
    )

    # Set file as dirty, save it and check that it has the right eol.
    codeeditor.document().setModified(True)
    editorstack.save()
    with open(fname, mode='r', newline='') as f:
        text = f.read()
    assert get_eol_chars(text) == get_eol_chars_from_os_name(os_name)


def test_save_with_os_eol_chars(editor_plugin, mocker, qtbot, tmpdir):
    """Check that saving new files uses eol chars according to OS."""
    editorstack = editor_plugin.get_current_editorstack()

    # Mock output of save file dialog.
    fname = osp.join(tmpdir, 'test_eol_chars.py')
    mocker.patch.object(editor_module, 'getsavefilename')
    editor_module.getsavefilename.return_value = (fname, '')

    # Load new, empty file
    editor_plugin.new()
    qtbot.wait(500)
    codeeditor = editor_plugin.get_current_editor()

    # Write some blank lines on it.
    for __ in range(3):
        qtbot.keyClick(codeeditor, Qt.Key_Return)

    # Save file and check that it has the right eol.
    editorstack.save()
    with open(fname, mode='r', newline='') as f:
        text = f.read()

    assert get_eol_chars(text) == os.linesep


def test_remove_editorstacks_and_windows(editor_plugin, qtbot):
    """
    Check that editor stacks and windows are removed from the list maintained
    for them in the editor when closing editor windows and split editors.

    This is a regression test for spyder-ide/spyder#20144.
    """
    # Create empty file
    editor_plugin.new()

    # Create editor window
    editor_window = editor_plugin.get_widget().create_new_window()
    qtbot.wait(500)  # To check visually that the window was created

    # This is not done automatically by Qt when running our tests (don't know
    # why), but it's done in normal usage. So we need to do it manually
    editor_window.editorwidget.editorstacks[0].deleteLater()

    # Close editor window
    editor_window.close()
    qtbot.wait(500)  # Wait for bit so window objects are actually deleted

    # Check the window objects were removed
    assert len(editor_plugin.get_widget().editorstacks) == 1
    assert len(editor_plugin.get_widget().editorwindows) == 0

    # Split editor and check the focus is given to the cloned editorstack
    editor_plugin.get_widget().editorsplitter.split()
    qtbot.wait(500)  # To check visually that the split was done
    assert editor_plugin.get_current_editor().is_cloned

    # Close editorstack
    editor_plugin.get_current_editorstack().close()
    qtbot.wait(500)  # Wait for bit so the editorstack is actually deleted
    assert len(editor_plugin.get_widget().editorstacks) == 1


def test_register_run_metadata(editor_plugin):
    """
    Check that run metadata is registered for Python files and deregistered for
    non-Python ones on renames.

    This is a regression test for spyder-ide/spyder#22630.
    """
    # Add run config for Python files
    widget = editor_plugin.get_widget()
    widget.supported_run_configurations = {
        "py": {RunContext.File, RunContext.Selection, RunContext.Cell}
    }

    # Create empty file
    editor_plugin.new()

    # Check the file was registered to be run
    editorstack = editor_plugin.get_current_editorstack()
    filename = editorstack.get_filenames()[0]
    assert filename in widget.file_per_id.values()

    # Rename file to a type that can't be run
    editor_plugin.renamed(filename, 'foo.md')

    # Check the file is no longer available to be run
    filename = editorstack.get_filenames()[0]
    assert filename not in widget.file_per_id.values()
    assert widget.file_per_id == {}


if __name__ == "__main__":
    pytest.main(['-x', osp.basename(__file__), '-vv', '-rw'])



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for editor calltips, hovers and hints."""

# Standard library imports
import os
import sys

# Third party imports
from qtpy.QtCore import Qt, QPoint
from qtpy.QtGui import QTextCursor
import pytest

# Local imports
from spyder.config.base import running_in_ci
from spyder.plugins.editor.extensions.closebrackets import (
    CloseBracketsExtension
)

# Constants
TEST_SIG = 'some_function(foo={}, hello=None)'
TEST_DOCSTRING = "This is the test docstring."
TEST_TEXT = """'''Testing something'''
def {SIG}:
    '''{DOC}'''


some_function""".format(SIG=TEST_SIG, DOC=TEST_DOCSTRING)


@pytest.mark.order(2)
def test_hide_calltip(completions_codeeditor, qtbot):
    """Test that calltips are hidden when a matching ')' is found."""
    code_editor, _ = completions_codeeditor
    code_editor.show()
    code_editor.raise_()
    code_editor.setFocus()

    text = 'a = "sometext {}"\nprint(a.format'
    # Set text to start
    code_editor.set_text(text)
    code_editor.go_to_line(2)
    code_editor.move_cursor(14)
    calltip = code_editor.calltip_widget
    assert not calltip.isVisible()

    with qtbot.waitSignal(code_editor.sig_signature_invoked, timeout=30000):
        qtbot.keyClicks(code_editor, '(', delay=3000)

    qtbot.waitUntil(lambda: calltip.isVisible(), timeout=3000)
    qtbot.keyClicks(code_editor, '"hello"')
    qtbot.keyClicks(code_editor, ')', delay=330)
    assert calltip.isVisible()
    qtbot.keyClicks(code_editor, ')', delay=330)
    qtbot.waitUntil(lambda: not calltip.isVisible(), timeout=3000)
    qtbot.keyClick(code_editor, Qt.Key_Enter, delay=330)
    assert not calltip.isVisible()


@pytest.mark.order(2)
@pytest.mark.parametrize('params', [
            # Parameter, Expected Output
            ('dict', 'dict'),
            ('type', 'type'),
            ('"".format', '-> str'),
            (TEST_TEXT, TEST_SIG)
        ]
    )
def test_get_calltips(qtbot, completions_codeeditor, params):
    """Test that the editor is returning hints."""
    code_editor, _ = completions_codeeditor

    param, expected_output_text = params

    # Set text in editor
    code_editor.set_text(param)

    # Move cursor to end of line
    code_editor.moveCursor(QTextCursor.End)

    code_editor.calltip_widget.hide()

    # Test with brackets autocompletion enable/disable
    bracket_extension = code_editor.editor_extensions.get(
        CloseBracketsExtension)

    # Bracket autocompletion enabled
    with qtbot.waitSignal(code_editor.sig_signature_invoked,
                          timeout=30000) as blocker:
        qtbot.keyPress(code_editor, Qt.Key_ParenLeft, delay=1000)

        # This is needed to leave time for the calltip to appear
        # and make the tests succeed
        qtbot.wait(2000)

        args = blocker.args
        print('args:', [args])
        output_text = args[0]['signatures']['label']
        assert expected_output_text in output_text
        code_editor.calltip_widget.hide()

    # Bracket autocompletion disabled
    bracket_extension.enable = False
    with qtbot.waitSignal(code_editor.sig_signature_invoked,
                          timeout=30000) as blocker:
        qtbot.keyPress(code_editor, Qt.Key_ParenLeft, delay=1000)

        # This is needed to leave time for the calltip to appear
        # and make the tests succeed
        qtbot.wait(2000)

        args = blocker.args
        print('args:', [args])
        output_text = args[0]['signatures']['label']
        assert expected_output_text in output_text
        code_editor.calltip_widget.hide()

    # Set bracket autocomplete to default value
    bracket_extension.enable = True


@pytest.mark.order(2)
@pytest.mark.skipif(not os.name == 'nt', reason='Only works on Windows')
@pytest.mark.parametrize('params', [
        # Parameter, Expected Output
        ('"".format', '-> str'),
        ('import math', 'module'),
        (TEST_TEXT, TEST_DOCSTRING)
    ]
)
def test_get_hints(qtbot, completions_codeeditor, params, capsys):
    """Test that the editor is returning hover hints."""
    code_editor, _ = completions_codeeditor
    param, expected_output_text = params

    # Move mouse to another position to be sure the hover is displayed when
    # the cursor is put on top of the tested word.
    qtbot.mouseMove(code_editor, QPoint(400, 400))

    # Set text in editor
    code_editor.set_text(param)

    # Get cursor coordinates
    code_editor.moveCursor(QTextCursor.End)
    qtbot.keyPress(code_editor, Qt.Key_Left)

    # Wait a bit in case the window manager repositions the window.
    qtbot.wait(1000)

    # Position cursor on top of the word we want the hover for.
    x, y = code_editor.get_coordinates('cursor')
    point = code_editor.calculate_real_position(QPoint(x, y))

    # Get hover and compare
    with qtbot.waitSignal(code_editor.sig_display_object_info,
                          timeout=30000) as blocker:
        qtbot.mouseMove(code_editor, point)
        qtbot.mouseClick(code_editor, Qt.LeftButton, pos=point)
        qtbot.waitUntil(lambda: code_editor.tooltip_widget.isVisible(),
                        timeout=10000)

        args = blocker.args
        print('args:', [args])
        output_text = args[0]
        assert expected_output_text in output_text
        code_editor.tooltip_widget.hide()


@pytest.mark.order(2)
@pytest.mark.skipif(sys.platform == 'darwin', reason='Fails on Mac')
@pytest.mark.parametrize('text', [
        'def test():\n    pass\n\ntest',
        '# a comment',
        '"a string"',
    ]
)
def test_get_hints_not_triggered(qtbot, completions_codeeditor, text):
    """Test that the editor is not returning hover hints for empty docs."""
    code_editor, _ = completions_codeeditor

    # Set text in editor
    code_editor.set_text(text)

    # Move mouse to another position.
    qtbot.mouseMove(code_editor, QPoint(400, 400))

    # Get cursor coordinates
    code_editor.moveCursor(QTextCursor.End)

    for _ in range(3):
        qtbot.keyPress(code_editor, Qt.Key_Left)

    # Wait a bit in case the window manager repositions the window.
    qtbot.wait(1000)

    # Position cursor on top of word we want the hover for.
    x, y = code_editor.get_coordinates('cursor')
    point = code_editor.calculate_real_position(QPoint(x, y))

    # Check that no hover was generated.
    with qtbot.waitSignal(code_editor.completions_response_signal,
                          timeout=30000):
        qtbot.mouseMove(code_editor, point)
        qtbot.mouseClick(code_editor, Qt.LeftButton, pos=point)
        qtbot.wait(1000)
        assert not code_editor.tooltip_widget.isVisible()


@pytest.mark.order(2)
@pytest.mark.skipif(
    (
        sys.platform == "darwin"
        or (sys.platform.startswith("linux") and running_in_ci())
    ),
    reason="Fails on CIs for Linux and Mac",
)
@pytest.mark.parametrize(
    "params",
    [
        ("an_int = 10", "This is an integer"),
        ("a_dict = {1: 2}", "This is a dictionary"),
        ('a_string = "foo"', "This is a string"),
    ],
)
def test_get_hints_for_builtins(qtbot, completions_codeeditor, params):
    """Test that the editor is showing the right hover hints for builtins."""
    code_editor, _ = completions_codeeditor
    text, expected = params

    # Set text in editor
    code_editor.set_text(text)

    # Move mouse to another position.
    qtbot.mouseMove(code_editor, QPoint(400, 400))

    # Move cursor at the beginning of the last line
    code_editor.moveCursor(QTextCursor.End)
    code_editor.moveCursor(QTextCursor.StartOfLine)

    for _ in range(3):
        qtbot.keyPress(code_editor, Qt.Key_Right)

    # Wait a bit in case the window manager repositions the window.
    qtbot.wait(1000)

    # Position cursor on top of the word we want the hover for.
    x, y = code_editor.get_coordinates('cursor')
    point = code_editor.calculate_real_position(QPoint(x, y))

    # Check the generated hover.
    with qtbot.waitSignal(
        code_editor.sig_display_object_info, timeout=30000
    ):
        # Give focus to the editor
        qtbot.mouseClick(code_editor, Qt.LeftButton, pos=point)

        # These two mouse moves are necessary to ensure the hover is shown.
        qtbot.mouseMove(code_editor, point)
        qtbot.mouseMove(
            code_editor,
            QPoint(point.x() + 2, point.y() + 2),
            delay=100
        )

        # Check hover is shown and contains the expected text
        qtbot.waitUntil(code_editor.tooltip_widget.isVisible, timeout=10000)
        assert expected in code_editor.tooltip_widget.text()


@pytest.mark.order(2)
@pytest.mark.skipif(
    (
        sys.platform == "darwin"
        or (sys.platform.startswith("linux") and running_in_ci())
    ),
    reason="Fails on CIs for Linux and Mac",
)
@pytest.mark.parametrize(
    "params",
    [
        ("import pandas as pd\npd.DataFrame", "bottom"),
        ("\n" * 15 + "import pandas as pd\npd.DataFrame", "top"),
    ],
)
def test_hints_vertical_position(qtbot, completions_codeeditor, params):
    """
    Test that we show hovers at the top or bottom of the text according to
    its length and text positioning in the editor.
    """
    code_editor, _ = completions_codeeditor
    text, expected = params

    # Set text in editor
    code_editor.set_text(text)

    # Move mouse to another position.
    qtbot.mouseMove(code_editor, QPoint(400, 400))

    # Move cursor at the beginning of the last line
    code_editor.moveCursor(QTextCursor.End)
    code_editor.moveCursor(QTextCursor.StartOfLine)

    for _ in range(4):
        qtbot.keyPress(code_editor, Qt.Key_Right)

    # Wait a bit in case the window manager repositions the window.
    qtbot.wait(1000)

    # Position cursor on top of the word we want the hover for.
    x, y = code_editor.get_coordinates('cursor')
    point = code_editor.calculate_real_position(QPoint(x, y))

    # Check the generated hover.
    with qtbot.waitSignal(
        code_editor.sig_display_object_info, timeout=30000
    ):
        # Give focus to the editor
        qtbot.mouseClick(code_editor, Qt.LeftButton, pos=point)

        # These two mouse moves are necessary to ensure the hover is shown.
        qtbot.mouseMove(code_editor, point)
        qtbot.mouseMove(
            code_editor,
            QPoint(point.x() + 2, point.y() + 2),
            delay=100
        )

        qtbot.waitUntil(code_editor.tooltip_widget.isVisible, timeout=10000)

        # Check hover is shown in the right position
        global_point = code_editor._calculate_position(at_point=point)
        if expected == "top":
            assert code_editor.tooltip_widget.pos().y() < global_point.y()
        else:
            assert code_editor.tooltip_widget.pos().y() > global_point.y()


@pytest.mark.order(2)
@pytest.mark.skipif(
    running_in_ci() and not os.name == 'nt',
    reason="Only works on Windows"
)
def test_completion_hints(qtbot, completions_codeeditor):
    """Test that the editor is returning hover hints."""
    code_editor, _ = completions_codeeditor
    completion_widget = code_editor.completion_widget
    tooltip_widget = code_editor.tooltip_widget

    # Move mouse to another position to be sure the hover is displayed when
    # the cursor is put on top of the tested word.
    qtbot.mouseMove(code_editor, QPoint(400, 400))

    # Set text in editor
    code_editor.set_text("import pandas as pd\npd.")

    # Wait a bit in case the window manager repositions the window.
    qtbot.wait(1000)

    # Trigger a code completion
    with qtbot.waitSignal(
        code_editor.completions_response_signal,
        timeout=30000
    ):
        # Get completions
        code_editor.moveCursor(QTextCursor.End)
        qtbot.keyPress(code_editor, Qt.Key_Tab)

    # Check the hint is shown and placed where it should be
    qtbot.waitUntil(tooltip_widget.isVisible)
    assert tooltip_widget.is_hint()
    assert (
        (completion_widget.x() + completion_widget.width() - 1)
        == tooltip_widget.x()
    )
    assert completion_widget.y() == tooltip_widget.y()

    # Move mouse to be on top of the hint
    x, y = code_editor.get_coordinates('cursor')
    point = code_editor.calculate_real_position(QPoint(x, y))
    hint_point = QPoint(
        point.x() + completion_widget.width() + 60,
        point.y() + 60
    )
    qtbot.mouseMove(code_editor, hint_point, delay=100)

    # Do a click on top of the hint and check it doesn't hide.
    with qtbot.waitSignal(
        tooltip_widget.sig_completion_help_requested,
        timeout=30000
    ):
        qtbot.mouseClick(tooltip_widget, Qt.LeftButton, pos=hint_point)

    qtbot.wait(200)
    assert tooltip_widget.isVisible()

    # Move cursor out of the hint and check it's still visible
    qtbot.mouseMove(
        code_editor,
        QPoint(hint_point.x(), hint_point.y() + tooltip_widget.height() + 20),
        delay=100
    )

    qtbot.wait(200)
    assert tooltip_widget.isVisible()

    # Move cursor to the line number area and check hint is still visible
    code_editor.moveCursor(QTextCursor.StartOfLine)
    x, y = code_editor.get_coordinates('cursor')
    point_1 = code_editor.calculate_real_position(QPoint(x, y))

    qtbot.mouseMove(
        code_editor,
        QPoint(
            point_1.x() - code_editor.linenumberarea.width() // 2, point_1.y()
        ),
        delay=100,
    )

    qtbot.wait(200)
    assert tooltip_widget.isVisible()

    # Hide completion widget and check hint is hidden too.
    completion_widget.hide()
    qtbot.waitUntil(lambda: not tooltip_widget.isVisible(), timeout=500)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Widget that handle communications between the IPython Console and
the Variable Explorer
"""

# Standard library imports
import logging
from pickle import PicklingError, UnpicklingError
import cloudpickle

# Third-party imports
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from spyder_kernels.comms.commbase import CommError

# Local imports
from spyder.config.base import _, is_conda_based_app

# For logging
logger = logging.getLogger(__name__)

# Max time before giving up when making a blocking call to the kernel
CALL_KERNEL_TIMEOUT = 30

# URL to our Github issues
GH_ISSUES = "https://github.com/spyder-ide/spyder/issues/new"


class NamepaceBrowserWidget(RichJupyterWidget):
    """
    Widget with the necessary attributes and methods to handle communications
    between the IPython Console and the kernel namespace
    """
    # --- Public API --------------------------------------------------
    def get_value(self, name):
        """Ask kernel for a value"""
        reason_big = _("The variable is too big to be retrieved")
        reason_not_picklable = _(
            "It was not possible to create a copy of the variable in the "
            "kernel to pass it to Spyder."
        )
        reason_dead = _("The kernel is dead")
        reason_other = _(
            "An unknown error occurred. Check the console because its contents "
            "could have been printed there."
        )
        reason_comm = _(
            "The channel used to communicate with the kernel is not working."
        )
        reason_missing_package_installer = _(
            "The '<tt>{}</tt>' module is required to open this variable. "
            "Unfortunately, it's not part of our installer, which means your "
            "variable can't be displayed by Spyder."
        )
        reason_missing_package = _(
            "The '<tt>{}</tt>' module is required to open this variable and "
            "it's not installed alongside Spyder. To fix this problem, please "
            "install it in the same environment that you use to run Spyder."
        )
        msg = _(
            "<br>%s<br><br>"
            "<b>Note</b>: If you consider this to be a valid error that needs "
            "to be fixed by the Spyder team, please report it on "
            "<a href='{}'>Github</a>."
        ).format(GH_ISSUES)

        try:
            value = self.call_kernel(
                blocking=True,
                display_error=True,
                timeout=CALL_KERNEL_TIMEOUT
            ).get_value(name, encoded=True)
            value = cloudpickle.loads(value)
            return value
        except TimeoutError:
            raise ValueError(msg % reason_big)
        except (PicklingError, UnpicklingError, TypeError):
            raise ValueError(msg % reason_not_picklable)
        except RuntimeError:
            raise ValueError(msg % reason_dead)
        except KeyError:
            raise
        except CommError:
            raise ValueError(msg % reason_comm)
        except ModuleNotFoundError as e:
            if is_conda_based_app():
                raise ValueError(
                    msg % reason_missing_package_installer.format(e.name)
                )
            else:
                raise ValueError(msg % reason_missing_package.format(e.name))
        except Exception:
            raise ValueError(msg % reason_other)

    def set_value(self, name, value):
        """Set value for a variable"""
        # Encode with cloudpickle and base64
        encoded_value = cloudpickle.dumps(value)
        self.call_kernel(
            interrupt=True,
            blocking=False,
            display_error=True,
            ).set_value(name, encoded_value, encoded=True)

    def remove_value(self, name):
        """Remove a variable"""
        self.call_kernel(
            interrupt=True,
            blocking=False,
            display_error=True,
            ).remove_value(name)

    def copy_value(self, orig_name, new_name):
        """Copy a variable"""
        self.call_kernel(
            interrupt=True,
            blocking=False,
            display_error=True,
            ).copy_value(orig_name, new_name)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# ----------------------------------------------------------------------------
"""Tests for plugin config dialog."""

from unittest.mock import MagicMock

# Test library imports
import pytest
from qtpy.QtWidgets import QMainWindow

# Local imports
from spyder.plugins.ipythonconsole.plugin import IPythonConsole
from spyder.plugins.preferences.tests.conftest import config_dialog


class MainWindowMock(QMainWindow):
    register_shortcut = MagicMock()
    editor = MagicMock()

    def __getattr__(self, attr):
        return MagicMock()


@pytest.mark.parametrize(
    'config_dialog',
    # [[MainWindowMock, [ConfigPlugins], [Plugins]]]
    [[MainWindowMock, [], [IPythonConsole]]],
    indirect=True)
def test_config_dialog(config_dialog):
    configpage = config_dialog.get_page()
    assert configpage
    configpage.save_to_conf()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder Run API."""

# Standard library imports
from __future__ import annotations

import functools
from datetime import datetime
import logging
from typing import Any, Callable, Set, List, Union, Optional, Dict, TypedDict

# Third-party imports
from qtpy.QtCore import QObject
from qtpy.QtWidgets import QWidget
from typing_extensions import NotRequired # Available from Python 3.11+

logger = logging.getLogger(__name__)


class RunActions:
    Run = 'run'
    Configure = 'configure'
    ReRun = 're-run last script'
    GlobalConfigurations = "global configurations"


class RunContextType(dict):
    def __init__(self, identifier: str):
        super().__init__()

    def __setattr__(self, key: str, value: str) -> None:
        if key not in self:
            self[key] = value

    def __getattribute__(self, key: str) -> str:
        if key in self:
            return self[key]
        return super().__getattribute__(key)


RunContext = RunContextType('context')
RunContext.File = 'file'
RunContext.Cell = 'cell'
RunContext.Selection = 'selection'

RunResultFormat = RunContextType('result display format')

RunResultFormat.NoDisplay = 'no_display'


class RunConfigurationMetadata(TypedDict):
    """Run input metadata schema."""

    # Human-readable name for the run input.
    name: str

    # Name of the RunConfigurationProvider that produced the run configuration.
    source: str

    # File path to the source used for the run configuration.
    path: str

    # Timestamp of the run input information.
    timestamp: datetime

    # Extra metadata information.
    extra: NotRequired[dict]

    # Unique identifier for this run configuration. This should be generated
    # and managed by the RunConfigurationProvider
    uuid: str

    # The context of the input provided. e.g. file, selection, cell or others.
    # This attribute can be customized by the `RunConfigurationProvider` when
    # registering with the run plugin. The context can be compared against the
    # values of `RunContext`. e.g. `info['context'] == RunContext.File`.
    context: Context

    # File extension or identifier of the input context.
    input_extension: str


class RunConfiguration(TypedDict):
    """Run input information schema."""

    # The output format to produce after executing the input. Each entry on the
    # set must belong to the `RunResultFormat` dict. The executor is
    # responsible of producing the correct format. This field will be available
    # on a RunExecutor but it is not necessary for a RunConfigurationProvider
    # to include it.
    output_formats: NotRequired[Set[str]]

    # Input to process by the executor. The executor is responsible for the
    # correct interpretation of the input type.
    run_input: Any

    # Run input metadata information.
    metadata: RunConfigurationMetadata


class RunExecutionMetadata(TypedDict):
    """Run result metadata schema."""

    # Human readable identifier for the run executor.
    executor: str

    # Start timestamp of the run execution.
    start_timestamp: datetime

    # End timestamp of the run execution.
    end_timestamp: datetime

    # Extra execution metadata. This field can be used to store resource
    # consumption, statistics and other fields that might be relevant.
    extra: NotRequired[dict]


class RunResultError(TypedDict):
    # Error code for the current error.
    code: Optional[int]

    # Human-readable message that describes the error.
    message: Optional[str]


class RunResult(TypedDict):
    """Run result entry schema."""

    # The output format for the current run result. It must belong to the
    # `RunResultDisplay` dict.
    output_format: str

    # The output of the run result. The viewer is responsible for the current
    # interpretation of the output type.
    run_output: Union[Any, RunResultError]

    # File extension or identifier of the input context.
    input_extension: str

    # Original run input metadata.
    metadata: RunConfigurationMetadata

    # Execution metadata.
    execution_metadata: RunExecutionMetadata

    # True if the execution result yielded an error, False otherwise. If the
    # execution failed, then the `run_output` field will comply with the
    # `RunResultError` schema.
    is_error: bool


PossibleRunResult = Union[RunResult, RunResultError]


class Context(TypedDict):
    """Run context name schema."""

    # CamelCase name of the context.
    name: str

    # String identifier for the run context. If non-existent or None, then a
    # snake_case version of the name is used.
    identifier: NotRequired[Optional[str]]


class OutputFormat(TypedDict):
    """Output format information schema."""

    # Human-readable name for the output format. It must be CamelCase
    name: str

    # String identifier for the output format. If non-existent or None, then a
    # snake_case version of the name is used.
    identifier: NotRequired[Optional[str]]


class ExtendedContext(TypedDict):
    """Extended context information schema."""

    # The specified context information.
    context: Context

    # True if entities identified with the given context can be registered as
    # run configurations via `register_run_configuration_metadata`. False if
    # the specified context is a subordinate of another one.
    is_super: bool


class SupportedExtensionContexts(TypedDict):
    """Supported file extension and contexts schema."""

    # File extension or identifier of the input context.
    input_extension: str

    # The supported contexts for the given input extension, e.g. file,
    # selection, cell or others.
    # The context can be compared against the values of `RunContext`. e.g.,
    # `info['context'] == RunContext.File`
    contexts: List[ExtendedContext]


class WorkingDirSource:
    ConfigurationDirectory = 0
    CurrentDirectory = 1
    CustomDirectory = 2


class WorkingDirOpts(TypedDict):
    """Run execution working directory options."""

    # Source used to look for the working directory that will be used in
    # the run execution.
    source: WorkingDirSource

    # If not None, then it specifies a path to a custom directory location.
    path: Optional[str]


class RunExecutionParameters(TypedDict):
    """Run execution parameters."""

    # The working directory options
    working_dir: WorkingDirOpts

    # The run executor parameters
    executor_params: dict


class ExtendedRunExecutionParameters(TypedDict):
    """Extended run execution parameters."""

    # The unique identifier for the execution parameter set.
    uuid: str

    # The name of the run execution parameter set.
    name: str

    # The run execution parameters.
    params: RunExecutionParameters

    # The unique identifier for the file to which these parameters correspond
    # to, if any.
    file_uuid: Optional[str]

    # To identify these parameters as default
    default: bool


class StoredRunExecutorParameters(TypedDict):
    """Per run executor configuration parameters."""

    # Dictionary that maps from parameter identifiers to the actual
    # configuration.
    params: Dict[str, ExtendedRunExecutionParameters]


class StoredRunConfigurationExecutor(TypedDict):
    """Stored run executor options per run configuration settings."""

    # Name of the last used run executor for the current run configuration.
    executor: Optional[str]

    # Unique identifier for the currently selected parameters. None
    # if using default or transient settings.
    selected: Optional[str]


class RunConfigurationProvider(QObject):
    """
    Interface used to retrieve inputs to run on a code executor.

    This API needs to be implemented by any plugin that wants to provide
    an input/file to a code runner, e.g. editor files to be executed in
    the IPythonConsole. It also needs to be covariant with respect to
    :class:`spyder.api.plugins.SpyderDockablePlugin`.
    """

    def get_run_configuration(self, uuid: str) -> RunConfiguration:
        """
        Return the run information for the specified identifier.

        Arguments
        ---------
        uuid: str
            The unique identifier for the requested run configuration. Such
            id should have been registered previously via
            `register_run_configuration_metadata` on the Run plugin.

        Returns
        -------
        configuration: RunConfiguration
            A dictionary containing the information required by the run
            executor.
        """
        raise NotImplementedError(f'{type(self)} must implement '
                                  'get_run_configuration')

    def get_run_configuration_per_context(
            self,
            context: str,
            extra_action_name: Optional[str] = None,
            context_modificator: Optional[str] = None,
            re_run: bool = False
        ) -> Optional[RunConfiguration]:
        """
        Return the run information for the given context.

        The run configuration requested must be returned based on the
        currently focused file/object/etc.

        Arguments
        ---------
        context: str
            The context identifier for which the run configuration
            is requested.
        extra_action_name: Optional[str]
            If not None, the name of the action that the provider should take
            after gathering the run configuration input. Else, no action needs
            to take place.
        context_modificator: Optional[str]
            str describing how to alter the context, e.g. run selection
            <from line>.
        re_run: bool
            If True, then the requested configuration should correspond to the
            last executed configuration for the given context.

        Returns
        -------
        configuration: Optional[RunConfiguration]
            A dictionary containing the information required by the run
            executor. If None, then the provider indicates to the Run plugin
            that the input needs to be discarded.
        """
        raise NotImplementedError(f'{type(self)} must implement '
                                  'get_run_configuration_per_context')

    def focus_run_configuration(self, uuid: str):
        """
        Switch the focus of the run configuration provider to
        the run configuration given by parameter.

        Arguments
        ---------
        uuid: str
            The unique identifier for the run configuration that should be
            focused on. Such id should have been registered previously via
            `register_run_configuration_metadata` on the Run plugin.
        """
        raise NotImplementedError(f'{type(self)} must implement '
                                  'focus_run_configuration')


RunExecuteFunc = Callable[
    [RunConfiguration, ExtendedRunExecutionParameters],
    List[PossibleRunResult]]


def run_execute(
        func: RunExecuteFunc = None,
        extension: Optional[Union[str, List[str]]] = None,
        context: Optional[Union[str, List[str]]] = None
    ) -> RunExecuteFunc:
    """
    Method decorator used to mark a method as an executor for a given file
    extension and context.

    The methods that use this decorator must have the following signature:

    .. code-block:: python
        def execution_handler(
            input: RunConfiguration,
            conf: ExtendedRunExecutionParameters) -> List[PossibleRunResult]:
            ...

    Arguments
    ---------
    func: RunExecuteFunc
        Method to mark as an executor handler. Given by default when applying
        the decorator.
    extension: Optional[Union[str, List[str]]]
        The file extension or list of file extensions that the executor
        should handle. If None then the method will handle all extensions.
    context: Optional[Union[str, List[str]]]
        The execution context or list of contexts that the executor should
        handle. If None then the method will handle all contexts.

    Returns
    -------
    func: RunExecuteFunc
        The same method that was given as input.

    Notes
    -----
    The method must not crash or raise an exception, instead the
    `RunResult` must wrap a `RunResultError` structure.
    """
    if func is None:
        return functools.partial(
            run_execute, extension=extension, context=context)

    if extension is None:
        extension = '__extension'

    if context is None:
        context = '__context'

    if isinstance(extension, str):
        extension = [extension]

    if isinstance(context, str):
        context = [context]

    run_exec_list = []
    for ext in extension:
        for ctx in context:
            run_exec_list.append((ext, ctx))
    func._run_exec = run_exec_list
    return func


class RunExecutor(QObject):
    """
    Interface used to execute run context information.

    This API needs to be implemented by any plugin that wants to execute
    an input produced by a :class:`RunConfigurationProvider` to produce an
    output compatible by a :class:`RunResultViewer`. It also needs to be
    covariant with respect to :class:`spyder.api.plugins.SpyderPluginV2`
    """

    def __init__(self, parent):
        super().__init__(parent)
        self._exec_methods: Dict[str, RunExecuteFunc] = {}
        self._exec_ext_methods: Dict[str, RunExecuteFunc] = {}
        self._exec_context_methods: Dict[str, RunExecuteFunc] = {}

        for method_name in dir(self):
            method = getattr(self, method_name, None)
            if hasattr(method, '_run_exec'):
                if not isinstance(method._run_exec, list):
                    continue
                for extension, context in method._run_exec:
                    if extension == '__extension' and context != '__context':
                        self._exec_ext_methods[context] = method
                    elif extension != '__extension' and context == '__context':
                        self._exec_context_methods[extension] = method
                    else:
                        self._exec_methods[(extension, context)] = method

    def exec_run_configuration(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[PossibleRunResult]:
        """
        Execute a run configuration.

        Arguments
        ---------
        input: RunConfiguration
            A dictionary containing the information required to execute the
            run input.
        conf: ExtendedRunExecutionParameters
            A dictionary containing the configuration parameters to use
            during execution.

        Returns
        -------
        results: List[PossibleRunResult]
            A list of `RunResult` dictionary entries, one per each `run_output`
            format requested by the input. Each entry must comply with the
            format requested.
        """
        metadata = input['metadata']
        context = metadata['context']
        extension = metadata['input_extension']
        context = RunContext[context['name']]
        query = (extension, context)

        logger.info('Running extension %s with context %s on executor %s',
                    extension, context, self)

        all_query = ('__extension', '__context')
        if query in self._exec_methods:
            method = self._exec_methods[query]
        elif context in self._exec_ext_methods:
            method = self._exec_ext_methods[context]
        elif extension in self._exec_context_methods:
            method = self._exec_context_methods[extension]
        elif all_query in self._exec_methods:
            method = self._exec_methods[all_query]
        else:
            raise NotImplementedError(
                f'There is no method available to '
                f'execute the requested context ({context}) and file '
                f'extension ({extension}) in {type(self)}')

        return method(input, conf)


class RunResultViewer:
    """
    Interface used to display run execution results.

    This API needs to be implemented by any plugin that wants to display
    an output produced by a :class:`RunResultViewer`. It also needs to be
    covariant with respect to :class:`spyder.api.plugins.SpyderPluginV2`
    """

    def display_run_result(self, result: RunResult):
        """
        Display the output of a run execution.

        Arguments
        ---------
        result: RunResult
            A `RunResult` entry that contains the run execution information,
            including both input and execution metadata, as well as the result
            representation in a format that the `RunResultViewer` instance
            supports.
        """
        raise NotImplementedError(
            f'{type(self)} must implement display_run_result')


class RunExecutorConfigurationGroup(QWidget):
    """
    QWidget subclass used to declare configuration options for a RunExecutor.

    Every executor that wants to add a configuration group to the Run
    dialog for a given context and input extension must subclass this
    interface.

    Parameters
    ---------
    parent: QWidget
        The instance of the run dialog, which will be always given by
        parameter.
    context: Context
        The run configuration context for which the dialog will provide
        options.
    input_extension: str
        The run configuration input extension for which the dialog will
        provide options.
    input_metadata: RunConfigurationMetadata
        The run configuration metadata that will be tentatively executed.

    Notes
    -----
    The aforementioned parameters will always be passed by the Run
    dialog instance, and no subclass should modify them.
    """

    def __init__(
        self,
        parent: QWidget,
        context: Context,
        input_extension: str,
        input_metadata: RunConfigurationMetadata
    ):
        """Create a run executor configuration widget."""
        super().__init__(parent)
        self.context = context
        self.input_extension = input_extension
        self.input_metadata = input_metadata

    def get_configuration(self) -> dict:
        """
        Obtain a dictionary containing the configuration options for
        the executor.
        """
        return {}

    @staticmethod
    def get_default_configuration() -> dict:
        """
        Obtain a dictionary containing the default values for the
        executor configuration.
        """
        return {}

    def set_configuration(self, config: dict):
        """
        Given a configuration dictionary compatible with the current
        configuration group, set the graphical elements to their corresponding
        values.
        """
        pass


RunExecutorConfigurationGroupFactory = Callable[
    [QWidget, Context, str, RunConfigurationMetadata],
    RunExecutorConfigurationGroup]

RunExecutorConfigurationGroupCreator = Union[
    RunExecutorConfigurationGroup, RunExecutorConfigurationGroupFactory]


class SupportedExecutionRunConfiguration(TypedDict):
    """Run execution configuration metadata."""

    # File extension or identifier of the input context.
    input_extension: str

    # The context for the given input extension, e.g. file, selection, cell or
    # others.
    # The context can be compared against the values of `RunContext`. e.g.,
    # `info['context'] == RunContext.File`
    context: Context

    # The output formats available for the given context and extension.
    output_formats: List[OutputFormat]

    # The configuration widget used to set the options for the current
    # input extension and context combination. If None or missing then no
    # configuration widget will be shown on the Run dialog for that
    # combination.
    configuration_widget: NotRequired[
        Optional[RunExecutorConfigurationGroupCreator]]

    # True if the executor requires a path in order to work. False otherwise
    requires_cwd: bool

    # Priority number used to select the executor. The lower, the better.
    priority: int



# %% first Code Cell's name
a = 10
"Hello"

# In[23]
li = [1, 2, 3]

# <codecell> Third Code Cell\'s name
import numpy as np
arr = np.array(li)

#%% Fourth Code/ Cell's % name\
s = """Z:\\escape\\test\\string\n"""
a



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder dialog widgets.
"""

# Third-party imports
from qtpy.QtCore import Qt
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QDialogButtonBox, QProxyStyle, QStyle

# Local imports
from spyder.utils.stylesheet import AppStyle


class _SpyderButtonsProxyStyle(QProxyStyle):
    """Style adjustments for SpyderDialogButtonBox."""

    def styleHint(self, hint, option=None, widget=None, return_data=None):
        if hint == QStyle.SH_DialogButtonLayout:
            # Use the Windows buttons layout to have a uniform layout in all
            # platforms. We selected that layout because Windows is our most
            # popular platform.
            # Solution found in https://stackoverflow.com/a/35907926/438386
            return QDialogButtonBox.WinLayout

        return super().styleHint(hint, option, widget, return_data)


class SpyderDialogButtonBox(QDialogButtonBox):
    """QDialogButtonBox widget for Spyder."""

    def __init__(self, buttons=None, orientation=Qt.Horizontal, parent=None):
        if buttons:
            super().__init__(buttons, orientation, parent)
        elif orientation:
            super().__init__(orientation=orientation, parent=parent)
        else:
            super().__init__(parent=parent)

        # Don't display icons on standard buttons. This is a problem on Linux
        button_constants = [
            QDialogButtonBox.Ok,
            QDialogButtonBox.Open,
            QDialogButtonBox.Save,
            QDialogButtonBox.Cancel,
            QDialogButtonBox.Close,
            QDialogButtonBox.Discard,
            QDialogButtonBox.Apply,
            QDialogButtonBox.Reset,
            QDialogButtonBox.RestoreDefaults,
            QDialogButtonBox.Help,
            QDialogButtonBox.SaveAll,
            QDialogButtonBox.Yes,
            QDialogButtonBox.YesToAll,
            QDialogButtonBox.No,
            QDialogButtonBox.NoToAll,
            QDialogButtonBox.Abort,
            QDialogButtonBox.Retry,
            QDialogButtonBox.Ignore,
        ]

        for constant in button_constants:
            button = self.button(constant)
            if button is not None:
                button.setIcon(QIcon())

        # Set a reasonable spacing between buttons. This is a problem on Mac
        self.layout().setSpacing(2 * AppStyle.MarginSize)

        # Set style
        style = _SpyderButtonsProxyStyle(None)
        style.setParent(self)
        self.setStyle(style)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
'''
Tests for editor panels.
'''

# Third party imports
from qtpy.QtCore import QSize, Qt
from qtpy.QtGui import QPainter, QColor, QFontMetrics
import pytest

# Local imports
from spyder.plugins.editor.api.panel import Panel


# --- External Example Panel
# -----------------------------------------------------------------------------
class EmojiPanel(Panel):
    """Example external panel."""

    def __init__(self):
        """Initialize panel."""
        Panel.__init__(self)
        self.setMouseTracking(True)
        self.scrollable = True

    def sizeHint(self):
        """Override Qt method.
        Returns the widget size hint (based on the editor font size).
        """
        fm = QFontMetrics(self.editor.font())
        size_hint = QSize(fm.height(), fm.height())
        if size_hint.width() > 16:
            size_hint.setWidth(16)
        return size_hint

    def _draw_red(self, top, painter):
        """Draw emojis.

        Arguments
        ---------
        top: int
            top of the line to draw the emoji
        painter: QPainter
            QPainter instance
        """
        painter.setPen(QColor('white'))
        font_height = self.editor.fontMetrics().height()
        painter.drawText(0, top, self.sizeHint().width(),
                         font_height, int(Qt.AlignRight | Qt.AlignBottom),
                         '👀')

    def paintEvent(self, event):
        """Override Qt method.
        Paint emojis.
        """
        super(EmojiPanel, self).paintEvent(event)
        painter = QPainter(self)
        painter.fillRect(event.rect(), self.editor.sideareas_color)
        for top, __, __ in self.editor.visible_blocks:
            self._draw_red(top, painter)


# --- Tests
# -----------------------------------------------------------------------------
@pytest.mark.parametrize('position', [
    Panel.Position.LEFT, Panel.Position.RIGHT, Panel.Position.TOP,
    Panel.Position.BOTTOM, Panel.Position.FLOATING])
def test_register_panel(setup_editor, position):
    """Test registering an example external panel in the editor."""
    editor_stack, editor = setup_editor

    # Register the panel
    editor_stack.register_panel(EmojiPanel, position=position)

    # Verify the panel is added in the panel manager
    new_panel = editor.panels.get(EmojiPanel)
    assert new_panel is not None

    # Verify the panel is in the editorstack
    assert (EmojiPanel, (), {}, position) in editor_stack.external_panels

    # Verify that the panel is shown in new files
    finfo = editor_stack.new('foo.py', 'utf-8', 'hola = 3\n')
    editor2 = finfo.editor

    new_panel = editor2.panels.get(EmojiPanel)
    assert new_panel is not None

    # Remove external panel
    editor_stack.external_panels = []
    editor.panels.remove(EmojiPanel)
    editor2.panels.remove(EmojiPanel)


if __name__ == '__main__':
    pytest.main()



# -*- coding: utf-8 -*-
"""
Spyder Editor

This is a temporary script file.
"""

import os
import sys


# %% functions
def d():
    def inner(): return 2
    return inner
# ---- func 1 and 2


def func1():
    for i in range(3):
        print(i)


def func2():
    if True:
        pass
# ---- other functions


def a():
    pass


def b():
    pass


def c():
    pass


# %% classes
class Class1:
    def __init__(self):
        super(Class1, self).__init__()
        self.x = 2

    def method3(self):
        pass

    def method2(self):
        pass

    def method1(self):
        pass



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import os
import logging
from packaging.version import parse

import pytest

from spyder.config.base import running_in_ci
from spyder.plugins.updatemanager import workers
from spyder.plugins.updatemanager.workers import (
    UpdateType, get_asset_info, WorkerUpdate, HTTP_ERROR_MSG
)
from spyder.plugins.updatemanager.widgets import update
from spyder.plugins.updatemanager.widgets.update import UpdateManagerWidget

logging.basicConfig()


@pytest.fixture(autouse=True)
def capture_logging(caplog):
    caplog.set_level(10, "spyder.plugins.updatemanager")


@pytest.fixture
def worker():
    return WorkerUpdate(None)


# ---- Test WorkerUpdate

@pytest.mark.parametrize("version", ["1.0.0", "1000.0.0"])
@pytest.mark.parametrize(
    "channel", [
        ("pkgs/main", "https://repo.anaconda.com/pkgs/main"),
        ("conda-forge", "https://conda.anaconda.org/conda-forge"),
    ]
)
def test_updates(qtbot, mocker, caplog, version, channel):
    """
    Test whether or not we offer updates according to the current Spyder
    version and package installation channel.

    Uses UpdateManagerWidget in order to also test QThread.
    """
    mocker.patch.object(update, "__version__", new=version)
    # Do not execute start_update after check_update completes.
    mocker.patch.object(
        UpdateManagerWidget, "start_update", new=lambda x: None
    )
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse(version))
    mocker.patch.object(
        workers, "get_spyder_conda_channel", return_value=channel
    )

    with caplog.at_level(logging.DEBUG, logger='spyder.plugins.updatemanager'):
        # Capture >=DEBUG logging messages for spyder.plugins.updatemanager
        # while checking for updates. Messages will be reported at the end
        # of the pytest run, and only if this test fails.
        um = UpdateManagerWidget(None)
        um.start_check_update()
        qtbot.waitUntil(um.update_thread.isFinished)

    if um.update_worker.error:
        # Possible 403 error - rate limit error, was encountered while doing
        # the tests
        # Check error message corresponds to the status code and exit early to
        # prevent failing the test
        assert um.update_worker.error == HTTP_ERROR_MSG.format(status_code="403")
        return

    assert not um.update_worker.error

    update_available = um.update_worker.update_available
    if version.split('.')[0] == '1':
        assert update_available
    else:
        assert not update_available


@pytest.mark.parametrize("release", ["6.0.0", "6.0.0b3"])
@pytest.mark.parametrize("version", ["4.0.0a1", "4.0.0"])
@pytest.mark.parametrize("stable_only", [True, False])
def test_update_non_stable(qtbot, mocker, version, release, stable_only):
    """Test we offer unstable updates."""
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse(version))

    release = parse(release)
    worker = WorkerUpdate(stable_only)
    worker._check_update_available([release])

    if release.is_prerelease and stable_only:
        assert not worker.update_available
    else:
        assert worker.update_available


@pytest.mark.parametrize("version", ["4.0.0", "6.0.0"])
def test_update_no_asset(qtbot, mocker, version):
    """Test update availability when asset is not available"""
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse(version))

    releases = [parse("6.0.1"), parse("6.100.0")]
    worker = WorkerUpdate(True)
    worker._check_update_available(releases)

    # For both values of version, there should be an update available
    # However, the available version should be 6.0.1, since there is
    # no asset for 6.100.0
    assert worker.update_available
    assert worker.latest_release == releases[0]


@pytest.mark.parametrize(
    "release,update_type",
    [
        ("6.0.1", UpdateType.Micro),
        ("6.1.0", UpdateType.Minor),
        ("7.0.0", UpdateType.Major)
    ]
)
@pytest.mark.parametrize("app", [True, False])
def test_get_asset_info(qtbot, mocker, release, update_type, app):
    mocker.patch.object(workers, "CURRENT_VERSION", new=parse("6.0.0"))
    mocker.patch.object(workers, "is_conda_based_app", return_value=app)

    info = get_asset_info(release)
    assert info['update_type'] == update_type

    if update_type == "major" or not app:
        assert info['url'].endswith(('.exe', '.pkg', '.sh'))
        assert info['filename'].endswith(('.exe', '.pkg', '.sh'))
    else:
        assert info['url'].endswith(".zip")
        assert info['filename'].endswith(".zip")


# ---- Test WorkerDownloadInstaller

@pytest.mark.skip(reason="Re-enable when alternate repo is available")
@pytest.mark.skipif(not running_in_ci(), reason="Download only in CI")
def test_download(qtbot, mocker):
    """
    Test download spyder installer.

    Uses UpdateManagerWidget in order to also test QThread.
    """
    um = UpdateManagerWidget(None)
    um.latest_release = "6.0.0a2"
    um._set_installer_path()

    # Do not execute _start_install after download completes.
    mocker.patch.object(
        UpdateManagerWidget, "_confirm_install", new=lambda x: None
    )

    um._start_download()
    qtbot.waitUntil(um.download_thread.isFinished, timeout=60000)

    assert os.path.exists(um.installer_path)


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-

# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

import pytest

from qtpy.QtCore import QObject, Signal

from spyder.plugins.completion.providers.fallback.provider import (
    FallbackProvider)
from spyder.plugins.completion.tests.conftest import qtbot_module


class CompletionManagerMock(QObject):
    sig_recv_tokens = Signal(list)

    def handle_response(self, client, req_id, response):
        tokens = list(response['params'])
        self.sig_recv_tokens.emit(list(tokens))


@pytest.fixture(scope='module')
def fallback_completions(qtbot_module, request):
    fallback = FallbackProvider(None, {})
    completions = CompletionManagerMock(None)

    with qtbot_module.waitSignal(fallback.sig_provider_ready, timeout=30000):
        fallback.start()

    def teardown():
        fallback.shutdown()

    request.addfinalizer(teardown)

    fallback.sig_response_ready.connect(completions.handle_response)
    return fallback, completions
